/**
  *  \file client/chart.cc
  *  \brief Starchart Drawing
  *
  *  This module contains the routines to draw the starchart. It is a
  *  more-or-less straightforward port of the PCC 1.0.17 starchart
  *  code.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <climits>
#include "client/chart.h"
#include "client/marks.h"
#include "cpluslib/bitops.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/clip.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "game/ufo.h"
#include "int/value.h"
#include "int/intglobal.h"
#include "int/data.h"
#include "game/pref.h"
#include "util/conf.h"
#include "game/team.h"
#include "client/labels.h"
#include "game/bbox.h"
#include "game/ion.h"
#include "util/misc.h"
#include "game/mine.h"
#include "game/univ.h"

#if 0
# include "gfx/gfxaa.h"
# define drawCircle drawCircleAA
# define drawLine   drawLineAA
#endif

static const int
    SCRingRadius     = 3,       ///< Radius of planet ring, in ly.
    SCMaxRingRadius  = 6,       ///< Max.\ radius of scaled planet ring, in pixels.
    SCCrossRadius    = 6,       ///< Size of starbase cross, in ly.
    SCMaxCrossRadius = 12,      ///< Max.\ radius of scaled SB cross, in pixels.
    SCObjCrossRadius = 5,       ///< Size of cross in center of a circular object (mine/UFO), in ly.
    SCMaxObjCrossRadius = 10,   ///< Max.\ size of scaled cross in center of a circular object, in pixels.
    SCWPCrossRadius  = 10,      ///< Size of cross at end of waypoint, in ly.
    SCMaxWPCrossRadius = 20;    ///< Max.\ size of scaled cross at end of waypoint, in pixels.

static const uint8 ionstorm_init[] = { 0x88, 0x00, 0x22, 0x00,
                                       0x88, 0x00, 0x22, 0x00 };
static const uint8 ionstorm_dense_init[] = { 0x88, 0x44, 0x22, 0x44,
                                             0x88, 0x44, 0x22, 0x44 };
static const uint8 ufofill_init[] = { 0x88, 0x55, 0x22, 0x55,
                                      0x88, 0x55, 0x22, 0x55 };

const GfxFillPattern IONSTORM_FILL       = ionstorm_init;
const GfxFillPattern IONSTORM_DENSE_FILL = ionstorm_dense_init;
const GfxFillPattern UFO_FILL            = ufofill_init;

/** Draw a, possibly filled, circle.
    \param f    zero for outline circle, nonzero for filled. */
static void
drawObject(GfxContext& ctx, GfxPoint pt, int r, int32 f)
{
    if (f) {
        drawFilledCircle(ctx, pt, r);
    } else {
        drawCircle(ctx, pt, r);
    }
}

/** Set fill style for drawing a minefield. Must frob the pattern a bit
    to make the minefield appear the same in all positions of the map. */
static void
setMineFillStyle(GfxContext& ctx, GfxPoint pt, GUniverse& /*univ*/, GMinefield& pmf)
{
    if (pmf.getOwner() == ::getPlayerId())
        /* avoid that own and foreign minefields 'overwrite' each other */
        ++pt.y;
    if (pmf.isWeb()) {
        ctx.setFillPattern(GRAY50_FILL);
        if ((pt.x + pt.y) & 1)
            ctx.getFillPattern().shiftUp(1);
    } else {
        ctx.setFillPattern(GRAY25_FILL);
        ctx.getFillPattern().shiftUp((15 + 2*(pt.x & 1) - pt.y) & 3);
    }
}

/***************************** GChartViewport ****************************/

/*! \class GChartViewport
    This encapsulates the logic for drawing a starchart. Starcharts
    can be displayed in various zoom levels specified as fractions. */

const atom_t GChartViewport::NO_FILTER;

/** Initialize. Call after setting the other parameters to compute
    derived information. */
void
GChartViewport::init()
{
    center_screen.x = view.x + view.w / 2;
    center_screen.y = view.y + view.h / 2;
    origin_screen.x = center_screen.x - scale(center.x);
    origin_screen.y = center_screen.y + scale(center.y);
}

/** Scale. Convert game distance into screen distance. The value n
    should be chosen so that the result is in range permissible for an
    int16 (because int may have 16 bits only, and drawing code might
    only be able to handle 16 bit coordinates). */
int
GChartViewport::scale(int n) const
{
    /* we want to round towards -INF instead of 0 */
    if ((-1)/2 == 0 && n < 0)
        return - (-long(n) * mult) / divi;
    return long(n) * mult / divi;
}

/** Scale, with Clipping. Compute scale(n)+add, but clip the result to
    be in the guaranteed range of int16. Unlike scale(), this is designed
    to handle out-of-range values of n. */
int
GChartViewport::scaleClip(int n, int add) const
{
    /* no need to fix rounding behaviour, this one's called to
       scale large values... */
    long l = long(n) * mult / divi + add;
    if (l < -0x7FFF)
        return -0x7FFF;
    else if (l > 0x7FFF)
        return 0x7FFF;
    else
        return l;
}

/** Unscale. Convert screen distance into game distance. */
int
GChartViewport::unscale(int n) const
{
    if ((-1)/2 == 0 && n < 0)
        return - (-long(n) * divi) / mult;
    return long(n) * divi / mult;
}

/** Draw Starchart, main entry point.
    \param can    Canvas to draw to
    \param opt    Options (what to draw)
    \param univ   Universe to display
    \param bbox   Applicable bounding box */
void
GChartViewport::draw(GfxCanvas& can, GChartOptions& opt, GUniverse& univ,
                     const GChartBBox& bbox)
{
    GfxClipRectFilter worker(view, can);

    init();
    // FIXME: set font....

    /* We now perform iteration over map images in the individual sub
       functions. This means there are no longer any overlap artifacts
       at the map borders. It comes at a hefty price, though: whereas
       a single 640x480 chart cost 2M cycles previously (most map
       images were clipped out early), it costs around 10M to draw all
       map images and clip away individual objects. We therefore
       prepare a list of all map images first, and only draw those,
       which gets us back into the old magnitudes.

       Measurement code using x86 timestamp counter is below; run on
       an otherwise idle machine.

       Each layer iterates through the rectangular images. This draws
       the map for regular and rectangle-wrap modes. For circular wrap,
       each layer must implement its own special-casing. */
    image_max = chart_conf.getNumRectangularImages();
    for (int i = 0; i < image_max; ++i) {
        /* A map image is visible if its bounding box has a nonempty
           intersection with our viewport. */
        GfxRect a;
        a.include(scaleXY(chart_conf.getSimplePointAlias(bbox.getMinimumXY(), i)));
        a.include(scaleXY(chart_conf.getSimplePointAlias(bbox.getMaximumXY(), i)));
        a.intersect(view);
        image_visible[i] = a.exists();
    }

    // int a, b, c;
    // asm("rdtsc" : "=eax"(a), "=edx"(c));

    drawAux(worker, opt, univ);

    // asm("rdtsc" : "=eax"(b), "=edx"(c));
    // printf("GChartViewport::draw -> %d cycles\n", b - a);
}

/** Get index of first visible map image.
    \pre draw must have been called
    \return index, or -1 if none */
inline int
GChartViewport::getFirstImage()
{
    return getNextImage(-1);
}

/** Get index of next map image after n.
    \return index, or -1 if none */
int
GChartViewport::getNextImage(int n)
{
    do
        ++n;
    while (n < image_max && !image_visible[n]);

    if (n < image_max)
        return n;
    else
        return -1;
}

/** Draw moving object.
    \param ctx     graphics context
    \param pt      center point, screen coordinates
    \param radius  radius, screen coordinates
    \param speed   warp factor
    \param heading heading angle, in degrees
    \param f       nonzero to fill the object (options.fill) */
void
GChartViewport::drawMovingObject(GfxContext& ctx,
                                 GfxPoint pt, int radius,
                                 int speed, int heading, int32 f)
{
    drawObject(ctx, pt, radius, f);
    if (speed > 0 && speed <= 20 && heading >= 0) {
        double h = heading * PI / 180;
        pt.x += roundToInt(radius * std::sin(h));
        pt.y -= roundToInt(radius * std::cos(h));

        int way = scale(speed * speed);
        int dx = roundToInt(way * std::sin(h));
        int dy = roundToInt(way * std::cos(h));

        drawArrow(ctx, pt, GfxPoint(pt.x + dx, pt.y - dy), 2*mult < divi ? 3 : 5);
    }
}

/** Draw starchart, helper routine. This does the actual drawing of
    one starchart image. */
void
GChartViewport::drawAux(GfxCanvas& can, GChartOptions& opt, GUniverse& univ)
{
    drawSectors(can, opt.show, opt.fill);
    if (opt.show & GChartOptions::co_Mine)
        drawMines(can, univ, opt.fill);
    if (opt.show & GChartOptions::co_Ufo)
        drawUfos(can, univ, opt.fill);
    if (opt.show & GChartOptions::co_Ion)
        drawIons(can, univ, opt.fill);
    if (opt.show & GChartOptions::co_Drawings)
        drawDrawings(can, univ, opt.show);
    drawShipSelAndVectors(can, univ, opt.show);
    drawPlanets(can, univ, opt.show);
    drawShips(can, univ, opt.show);
}

/** Draw a cross, used for the center of Ufos and mine fields. */
void
GChartViewport::drawCross(GfxContext& ctx, GfxPoint pt)
{
    int n = scale(SCObjCrossRadius);
    if (n <= 0)
        n = 1;
    else if (n > SCMaxObjCrossRadius)
        n = SCMaxObjCrossRadius;
    drawHLine(ctx, pt.x - n, pt.y, pt.x + n);
    drawVLine(ctx, pt.x, pt.y - n, pt.y + n);
}

/** Draw ion storms. Call only if ion storm drawing is actually enabled.
    \param can canvas to use
    \param univ universe
    \param fill options.fill */
void
GChartViewport::drawIons(GfxCanvas& can, GUniverse& univ, int fill)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setFillPattern(IONSTORM_FILL);

    for (int sid = univ.ty_ion_storms.findNextIndex(0); sid != 0; sid = univ.ty_ion_storms.findNextIndex(sid)) {
        const GIonStorm& i = univ.getIonStorm(sid);
        if (i.getVoltage() < 50)
            ctx.setColor(COLOR_BLUE);
        else if (i.getVoltage() < 100)
            ctx.setColor(COLOR_GRAY);
        else if (i.getVoltage() < 150)
            ctx.setColor(COLOR_WHITE);
        else
            ctx.setColor(COLOR_RED);

        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            drawMovingObject(ctx,
                             scaleXY(chart_conf.getSimplePointAlias(i.getPos(), img)),
                             scale(i.getRadius()),
                             i.getSpeed(), i.getHeading(),
                             fill & GChartOptions::co_Ion);
        }
    }
}

/** Draw single user drawing.
    \param can  Canvas to draw on
    \param d    Drawing to display
    \param color Color slot to use, overriding marker's color
    \param with_text Flag saying whether we want marker labels displayed */
void
GChartViewport::drawDrawing(GfxCanvas& can, const GDrawing& d, uint8 color, bool with_text)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setFillPattern(IONSTORM_FILL);
    ctx.setTextAlign(1, 0);
    ctx.useFont(*fonts[getFont()]);
    ctx.setColor(color);

    switch (d.getType()) {
     case LineDrawing:
        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            GfxPoint origin = scaleXY(chart_conf.getSimplePointAlias(d.getPos(),  img));
            GfxPoint end    = scaleXY(chart_conf.getSimplePointAlias(d.getPos2(), img));
            drawLine(ctx, origin, end);
        }
        break;
     case RectangleDrawing:
        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            GfxPoint origin = scaleXY(chart_conf.getSimplePointAlias(d.getPos(),  img));
            GfxPoint end    = scaleXY(chart_conf.getSimplePointAlias(d.getPos2(), img));
            drawRectangle(ctx, GfxRect(std::min(origin.x, end.x),
                                       std::min(origin.y, end.y),
                                       std::abs(origin.x - end.x) + 1,
                                       std::abs(origin.y - end.y) + 1));
        }
        break;
     case CircleDrawing:
        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            GfxPoint origin = scaleXY(chart_conf.getSimplePointAlias(d.getPos(), img));
            drawCircle(ctx, origin, scale(d.getCircleRadius()));
        }
        break;
     case MarkerDrawing:
        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            GfxPoint origin = scaleXY(chart_conf.getSimplePointAlias(d.getPos(), img));
            if (divi < 2*mult) {
                /* draw marker */
                const TMarkerImage& img = getUserMarker(d.getMarkerKind(), true);
                drawMarker(ctx, img, origin);

                /* draw text */
                if (with_text) {
                    string_t text = strFirst(d.getComment(), "|");
                    if (text.size())
                        outTextF(ctx, origin.x, origin.y + img.height, 600, text);
                }
            } else {
                drawMarker(ctx, getUserMarker(d.getMarkerKind(), false), origin);
            }
        }
        break;
    }
}

/** Draw scanner beam.
    \param a,b Points
    \param color Color to use */
void
GChartViewport::drawScanner(GfxContext& ctx, GPoint a, GPoint b)
{
    GfxPoint aa = scaleXY(a);
    GfxPoint bb = scaleXY(b);

    drawScannerEnd(ctx, aa);
    if (aa != bb) {
        drawLine(ctx, aa, bb);
        drawScannerEnd(ctx, bb);
    }
}

void
GChartViewport::drawScannerEnd(GfxContext& ctx, GfxPoint pt)
{
    if (mult >= divi) {
        drawVLine(ctx, pt.x-3, pt.y-2, pt.y+2);
        drawVLine(ctx, pt.x+3, pt.y-2, pt.y+2);
        drawHLine(ctx, pt.x-2, pt.y-3, pt.x+2);
        drawHLine(ctx, pt.x-2, pt.y+3, pt.x+2);
    } else {
        drawPixel(ctx, GfxPoint(pt.x+1, pt.y));
        drawPixel(ctx, GfxPoint(pt.x-1, pt.y));
        drawPixel(ctx, GfxPoint(pt.x, pt.y+1));
        drawPixel(ctx, GfxPoint(pt.x, pt.y-1));
    }
}

/** Draw user drawings. Call only if these are actually enabled.
    \param can  Canvas to draw on
    \param trn  Turn to get drawings from
    \param show options.show, to test whether labels should be drawn */
void
GChartViewport::drawDrawings(GfxCanvas& can, GUniverse& trn, int show)
{
    GDrawingContainer& d = trn.getDrawings();
    const bool with_text = (show & GChartOptions::co_Labels) != 0;
    for (GDrawingContainer::iterator i = d.begin(); i != d.end(); ++i) {
        if ((*i)->isVisible()
            && (drawing_filter == NO_FILTER
                || (*i)->getTag() == drawing_filter))
        {
            drawDrawing(can, **i, (*i)->getColor(), with_text);
        }
    }

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    GExplosionContainer& e = trn.getExplosions();
    for (GExplosionContainer::iterator i = e.begin(); i != e.end(); ++i) {
        bool big = divi < 2*mult;
        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            GfxPoint pos = scaleXY(chart_conf.getSimplePointAlias(i->getPos(), img));
            /* red '+' */
            ctx.setColor(COLOR_RED);
            drawMarker(ctx, getUserMarker(0, big), pos);
            /* yellow 'x' */
            ctx.setColor(COLOR_YELLOW);
            drawMarker(ctx, getUserMarker(2, big), pos);
        }
    }
}

/** Draw Minefields. Call only if minefield drawing is actually enabled.
    \param can   canvas to draw to
    \param univ   turn to display
    \param fill  options.fill */
void
GChartViewport::drawMines(GfxCanvas& can, GUniverse& univ, int fill)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    const GMinefieldType& mines = univ.ty_minefields;
    for (GMinefieldType::const_iterator i = mines.begin(); i != mines.end(); ++i) {
        GMinefield& pmf = *i;
        setMineFillStyle(ctx, origin_screen, univ, pmf);

        GPlayerRelation rel = getPlayerRelation(pmf.getOwner());
        if (rel == is_Me)
            ctx.setColor(COLOR_GREENSCALE+4);
        else if (rel == is_Ally)
            ctx.setColor(COLOR_DARKYELLOWSCALE+4);
        else
            ctx.setColor(COLOR_DARK);

        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            GfxPoint pt = scaleXY(chart_conf.getSimplePointAlias(pmf.getPos(), img));
            drawObject(ctx, pt, scale(pmf.getRadius()), fill & GChartOptions::co_Mine);
            drawCross(ctx, pt);
        }
    }
}

void
GChartViewport::drawUfos(GfxCanvas& can, GUniverse& trn, int fill)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    const GUfoType& uc = trn.ty_ufos;
    for (GUfoType::const_iterator i = uc.begin(); i != uc.end(); ++i) {
        const GUfo& ufo = *i;
        if (ufo.isExistant()) {
            for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
                /* set draw style */
                GfxPoint pt = scaleXY(chart_conf.getSimplePointAlias(ufo.getPos(), img));
                ctx.setFillPattern(UFO_FILL)
                   .getFillPattern().shiftDown(pt.y & 3)
                                    .shiftRight(pt.x & 3);
                ctx.setColor(ufo.getColor());

                /* draw it */
                drawMovingObject(ctx, pt,
                                 scale(ufo.getRadius()),
                                 ufo.getSpeed(),
                                 ufo.getHeading(),
                                 fill & GChartOptions::co_Ufo);
                drawCross(ctx, pt);

                /* draw other end */
                if (GUfo* oe = ufo.getOtherEnd()) {
                    /* The "other end" relation is now symmetric, so
                       make sure to draw it only once. */
                    GPoint thisPos = ufo.getPos();
                    GPoint otherPos = oe->getPos();
                    if (thisPos.y < otherPos.y || (thisPos.y == otherPos.y && thisPos.x < otherPos.x)) {
                        GfxPoint pt2 = scaleXY(chart_conf.getSimplePointAlias(otherPos.getSimpleNearestAlias(thisPos), img));
                        drawLine(ctx, pt, pt2);
                    }
                }
            }
        }
    }
}

/** Draw sector/starchart borders.
    \param can    output canvas
    \param show   options.show */
void
GChartViewport::drawSectors(GfxCanvas& can, int show, int fill)
{
    GfxContext ctx;
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useCanvas(can);
    ctx.setColor(COLOR_DARK);

    if (chart_conf.getMode() != GChartConfiguration::Circular) {
        // Rectangular wrap or flat universe
        int dx = std::min(int(chart_conf.getSize().x / 200), 10);
        int dy = std::min(int(chart_conf.getSize().y / 200), 10);
        if (show & GChartOptions::co_Sectors) {
            for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
                GPoint center = chart_conf.getSimplePointAlias(chart_conf.getCenter(), img);

                ctx.setLinePattern((origin_screen.y & 1) ? 0x55 : 0xAA);
                for (int i = -dx; i <= dx; ++i)
                    drawVLine(ctx,
                              scaleX(center.x + 100*i),
                              scaleY(center.y - 100*dy),
                              scaleY(center.y + 100*dy));

                ctx.setLinePattern((origin_screen.x & 1) ? 0x55 : 0xAA);
                for (int i = -dy; i <= dy; ++i)
                    drawHLine(ctx,
                              scaleX(center.x - 100*dx),
                              scaleY(center.y + 100*i),
                              scaleX(center.x + 100*dx));
            }
        }

        if (show & GChartOptions::co_Borders) {
            for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
                GPoint p1 = chart_conf.getSimplePointAlias(chart_conf.getMinimumCoordinates(), img);
                GPoint p2 = chart_conf.getSimplePointAlias(chart_conf.getMaximumCoordinates(), img);

                ctx.setLinePattern(ror8(0x27, origin_screen.y));
                drawVLine(ctx, scaleX(p1.x), scaleY(p1.y), scaleY(p2.y));
                drawVLine(ctx, scaleX(p2.x), scaleY(p1.y), scaleY(p2.y));

                ctx.setLinePattern(ror8(0x27, origin_screen.x));
                drawHLine(ctx, scaleX(p1.x), scaleY(p1.y), scaleX(p2.x));
                drawHLine(ctx, scaleX(p1.x), scaleY(p2.y), scaleX(p2.x));
            }
        }
    } else {
        // Spherical wrap is special
        const int size = chart_conf.getSize().x;
        const int dx = std::min(int(size / 100), 10);
        const int dy = dx;
        const GPoint center = chart_conf.getCenter();
        if (show & GChartOptions::co_Sectors) {
            // Inside lines
            int32_t sr = square(size);

            ctx.setLinePattern((origin_screen.y & 1) ? 0x55 : 0xAA);
            for (int i = -dx; i <= dx; ++i) {
                int yc = std::min(int(std::sqrt(double(sr - square(100 * i)))), 1000);
                drawVLine(ctx, scaleX(center.x + 100*i), scaleY(center.y - yc), scaleY(center.y + yc));
            }

            ctx.setLinePattern((origin_screen.x & 1) ? 0x55 : 0xAA);
            for (int i = -dy; i <= dy; ++i) {
                int xc = std::min(int(std::sqrt(double(sr - square(100 * i)))), 1000);
                drawHLine(ctx, scaleX(center.x - xc), scaleY(center.y - 100*i), scaleX(center.x + xc));
            }
        }

        if (show & GChartOptions::co_Borders) {
            // Border
            ctx.setLinePattern(0x27);
            drawCircle(ctx, scaleXY(center), scale(size));
        }

        if (show & ~fill & GChartOptions::co_Sectors) {
            // Vertical outside lines
            ctx.setLinePattern(0x55);
            for (int xi = -dx; xi <= dx; ++xi) {
                bool drawing = false;
                GfxPoint cursor;
                for (int yi = -10*dy; yi <= 10*dy; ++yi) {
                    GPoint pt(100*xi + center.x, 10*yi + center.y);
                    GPoint pt1;
                    if (chart_conf.getPointAlias(pt, pt1, 1, false)) {
                        GfxPoint c = scaleXY(pt1);
                        if (drawing) {
                            drawLine(ctx, cursor, c);
                        }
                        cursor = c;
                        drawing = true;
                    } else {
                        drawing = false;
                    }
                }
            }

            // Horizontal outside lines
            for (int yi = -dy; yi <= dy; ++yi) {
                bool drawing = false;
                GfxPoint cursor;
                for (int xi = -10*dx; xi <= 10*dx; ++xi) {
                    GPoint pt(10*xi + center.x, 100*yi + center.y);
                    GPoint pt1;
                    if (chart_conf.getPointAlias(pt, pt1, 1, false)) {
                        GfxPoint c = scaleXY(pt1);
                        if (drawing) {
                            drawLine(ctx, cursor, c);
                        }
                        cursor = c;
                        drawing = true;
                    } else {
                        drawing = false;
                    }
                }
            }
        }
    }
}

/** Draw ship trail.
    \param ctx Context
    \param sh Ship to draw
    \param color Base color to use; use [color,color+8) */
void
GChartViewport::drawShipTrail(GfxContext& ctx, const GShip& sh, int turn, int color, int img)
{
    /* Implementation note: we do not assume any knowledge about how many
       ship track entries a ship contains. Since we're allowed to use 8 colors,
       we simply try to draw 16 entries. */
    enum { N = 16 };
    turn -= N;
    for (int i = 0; i < N; ++i, ++turn) {
        const TDbShipTrackEntry& me = sh.getHistoryEntry(turn);
        const TDbShipTrackEntry& next = sh.getHistoryEntry(turn+1);
        ctx.setColor(color + (i>>1));

        /* Draw knob here */
        if (me.x > 0) {
            GfxCanvas& can = ctx.getCanvas();
            GfxPoint mepos = scaleXY(chart_conf.getSimplePointAlias(GPoint(me.x, me.y), img));
            can.drawPixel(GfxPoint(mepos.x, mepos.y-1), ctx.getRawColor(), GFX_ALPHA_OPAQUE);
            can.drawPixel(GfxPoint(mepos.x-1, mepos.y), ctx.getRawColor(), GFX_ALPHA_OPAQUE);
            can.drawPixel(GfxPoint(mepos.x+1, mepos.y), ctx.getRawColor(), GFX_ALPHA_OPAQUE);
            can.drawPixel(GfxPoint(mepos.x, mepos.y+1), ctx.getRawColor(), GFX_ALPHA_OPAQUE);
        }

        /* Draw lines */
        if (me.x > 0 && next.x > 0) {
            /* Both positions known, so simply connect them */
            // FIXME: wrapping code is missing here
//         IF ChartConf.Wrap=cmWrap THEN BEGIN
//           sx := ScaleL(view, ChartConf.SizeX);
//           sy := ScaleL(view, ChartConf.SizeY);
//           IF nX+sx < XCor THEN Inc(nX, 2*sx) ELSE
//           IF Xcor < nX-sx THEN Dec(nX, 2*sx);

//           IF nY+sy < YCor THEN Inc(nY, 2*sy) ELSE
//           IF Ycor < nY-sy THEN Dec(nY, 2*sy);
//         {$IFDEF CircularWrap}
//         END ELSE IF ChartConf.Wrap=cmCircle THEN BEGIN
//           { check whether we wrapped this turn }
//           sx := psh^.locs[slot+1].X - psh^.locs[slot].X;
//           sy := psh^.locs[slot+1].Y - psh^.locs[slot].Y;
//           IF Sqr(LONGINT(sx)) + Sqr(LONGINT(sy)) >= Sqr(LONGINT(ChartConf.SizeX)) THEN BEGIN
//             { yes, we wrapped.
//               Draw two lines: one from the previous location to the
//               current point outside, one from the previous outside
//               point to the current location }
//             sx := psh^.locs[slot].X;
//             sy := psh^.locs[slot].Y;
//             qx := psh^.locs[slot+1].X;
//             qy := psh^.locs[slot+1].Y;
//             IF CircularMoveOutside(sx, sy, TRUE) AND CircularMoveOutside(qx, qy, TRUE) THEN BEGIN
//               MLineTo(view.x0 + ScaleL(view, qx), view.y0 - ScaleL(view, qy));
//               MoveTo(view.x0 + ScaleL(view, sx), view.y0 - ScaleL(view, sy));
//             END;
//           END;
//         {$ENDIF}
            GfxPoint mepos   = scaleXY(chart_conf.getSimplePointAlias(GPoint(me.x, me.y).getSimpleNearestAlias(GPoint(next.x, next.y)), img));
            GfxPoint nextpos = scaleXY(chart_conf.getSimplePointAlias(GPoint(next.x, next.y), img));
            drawLine(ctx, mepos, nextpos);
        } else if (me.x > 0 && me.heading >= 0 && me.speed > 0) {
            /* This position and heading known, so draw a line leaving here */
            int way = square(me.speed) / 2;
            if (way < 15)
                way = 15;
            GfxPoint mepos   = scaleXY(chart_conf.getSimplePointAlias(GPoint(me.x, me.y), img));
            GfxPoint nextpos = scaleXY(chart_conf.getSimplePointAlias(GPoint(me.x + roundToInt(way * std::sin(me.heading * PI / 180)),
                                                                             me.y + roundToInt(way * std::cos(me.heading * PI / 180))),
                                                                      img));
            drawLine(ctx, mepos, nextpos);
        } else if (next.x > 0 && next.heading >= 0 && next.speed > 0) {
            /* This position and heading known, so draw a line arriving there */
            int way = square(next.speed) / 2;
            if (way < 15)
                way = 15;
            GfxPoint nextpos = scaleXY(chart_conf.getSimplePointAlias(GPoint(next.x, next.y), img));
            GfxPoint mepos   = scaleXY(chart_conf.getSimplePointAlias(GPoint(next.x - roundToInt(way * std::sin(next.heading * PI / 180)),
                                                                       next.y - roundToInt(way * std::cos(next.heading * PI / 180))),
                                                                img));
            drawLine(ctx, mepos, nextpos);
        }
    }
}

void
GChartViewport::drawShipVector(GfxCanvas& can, const GShip& sh, int turn, int img)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    int arrowHeadSize = 2*mult > divi ? 5 : 3;
    int waypointCrossSize  = scale(SCWPCrossRadius);
    if (waypointCrossSize > SCMaxWPCrossRadius)
        waypointCrossSize = SCMaxWPCrossRadius;

    /* Trail */
    const GPlayerRelation rel = getPlayerRelation(sh.getOwner());
    drawShipTrail(ctx, sh, turn, (rel == is_Me ? COLOR_GREENSCALE
                                  : rel == is_Ally ? COLOR_DARKYELLOWSCALE
                                  : COLOR_FIRE),
                  img);

    /* Auto Task */
//     ppred := Ships^[sid].Predict;
//     IF ppred<>NIL THEN BEGIN
//       Color := Dark;
//       WriteMode := wmOR;
//       x0 := Ships^[sid].x;
//       y0 := Ships^[sid].y;
//       x0t := view.x0 + ScaleL(view, x0);
//       y0t := view.y0 - ScaleL(view, y0);
//       FOR i:=1 TO ppred^.valid DO BEGIN
//         x1 := ppred^.x[i];
//         y1 := ppred^.y[i];
//         MoveToNearest(x0,y0,x1,y1);
//         x1t := view.x0 + ScaleL(view, x1);
//         y1t := view.y0 - ScaleL(view, y1);
//         LongLine(x0t, y0t, x1t-x0t, y1t-y0t);
//         IF (x1t > -waypointCrossSize) AND (y1t > -waypointCrossSize) AND (x1t < 5000) AND (y1t < 5000) THEN BEGIN
//           j := waypointCrossSize DIV 2;
//           HLine(x1t-j, y1t, x1t+j);
//           VLine(x1t, y1t-j, y1t+j);
//         END;
//         x0 := x1;
//         y0 := y1;
//         x0t := x1t;
//         y0t := y1t;
//       END;
//       WriteMode := 0;
//     END;

    /* Heading */
    if (sh.getShipKind() == GShip::CurrentShip) {
        /* It's a current ship, we know its waypoint */
        int dx = sh.getWaypointDX();
        int dy = sh.getWaypointDY();
        GPoint pos = chart_conf.getSimplePointAlias(sh.getPos(), img);
        if (dx != 0 || dy != 0) {
            ctx.setColor(COLOR_DARK);
            GfxPoint a = scaleXY(pos);
            GfxPoint b = scaleXY(GPoint(pos.x + dx, pos.y + dy));
            drawLine(ctx, a, b);
            drawHLine(ctx, b.x-waypointCrossSize, b.y, b.x+waypointCrossSize);
            drawVLine(ctx, b.x, b.y-waypointCrossSize, b.y+waypointCrossSize);

            /* FIXME: it would be a good thing[tm] to use actual movement formulas here.
               Now, we're using the rough approximation also used in 1.x */
            if (sh.getWarp() > 0) {
                double factor = square(sh.getWarp()) / std::sqrt(double(square(dx) + square(dy)));
                if (sh.canDoSpecial(hf_Gravitonic))
                    factor *= 2;
                GfxPoint c = scaleXY(GPoint(pos.x + roundToInt(factor*dx), pos.y + roundToInt(factor*dy)));
                ctx.setColor(COLOR_GRAY);
                drawArrow(ctx, a, c, arrowHeadSize);
            }
        }
    } else if (sh.getHeading().isKnown() && sh.getWarp().isKnown()) {
        /* It's a target, we know its heading and speed */
        const int16 heading = sh.getHeading();
        const int16 speed   = sh.getWarp();
        if (speed > 0) {
            int32 dist = square(speed);
            if (sh.canDoSpecial(hf_Gravitonic))
                dist *= 2;
            ctx.setColor(COLOR_GRAY);
            GPoint pos = chart_conf.getSimplePointAlias(sh.getPos(), img);
            GfxPoint origin = scaleXY(pos);
            GfxPoint end = scaleXY(GPoint(pos.x + roundToInt(dist * std::sin(heading * PI / 180)),
                                          pos.y + roundToInt(dist * std::cos(heading * PI / 180))));
            drawArrow(ctx, origin, end, arrowHeadSize);
        }
    }
}

/** Draw Ships, part I. Draws ship selection and trails. These are
    drawn below the planets, to not obscure other stuff. */
void
GChartViewport::drawShipSelAndVectors(GfxCanvas& can, GUniverse& univ, int show)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    if (show & GChartOptions::co_Selection) {
        // Regular images
        ctx.setColor(COLOR_YELLOW);
        for (int i = univ.ty_any_ships.findNextIndexNoWrap(0, true); i != 0; i = univ.ty_any_ships.findNextIndexNoWrap(i, true)) {
            GShip& s = univ.getShip(i);
            for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
                GfxPoint pt = scaleXY(chart_conf.getSimplePointAlias(s.getPos(), img));
                drawSelection(ctx, pt, mult, divi);
            }
        }

        // Special case for circular wrap
        if (chart_conf.getMode() == GChartConfiguration::Circular) {
            for (int i = univ.ty_any_ships.findNextIndexNoWrap(0, true); i != 0; i = univ.ty_any_ships.findNextIndexNoWrap(i, true)) {
                GShip& s = univ.getShip(i);
                GPoint pt;
                if (chart_conf.getPointAlias(s.getPos(), pt, 1, true)) {
                    drawSelection(ctx, scaleXY(pt), mult, divi);
                }
            }
        }
    }

    if (show & GChartOptions::co_Messages) {
        ctx.setColor(COLOR_BRIGHTBLUE);
        ctx.setColor(COLOR_BRIGHTPINK);
        ctx.setColor(COLOR_BRIGHTMAGENTA);
        for (int i = univ.ty_any_ships.findNextIndexNoWrap(0, false); i != 0; i = univ.ty_any_ships.findNextIndexNoWrap(i, false)) {
            GShip& s = univ.getShip(i);
            if (!s.getAssociatedMessages().empty()) {
                for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
                    GfxPoint pt = scaleXY(chart_conf.getSimplePointAlias(s.getPos(), img));
                    drawFlag(ctx, pt, mult, divi);
                }
            }
        }
        for (int i = univ.ty_any_planets.findNextIndexNoWrap(0, false); i != 0; i = univ.ty_any_planets.findNextIndexNoWrap(i, false)) {
            GPlanet& s = univ.getPlanet(i);
            if (!s.getAssociatedMessages().empty()) {
                for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
                    GfxPoint pt = scaleXY(chart_conf.getSimplePointAlias(s.getPos(), img));
                    drawFlag(ctx, pt, mult, divi);
                }
            }
        }
    }

    if (!(show & GChartOptions::co_ShipDots)) {
        const bool big = (divi < 2*mult);
        for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
            const GShip& s = univ.getShip(i);
            if (s.getOrbitPlanetId() == 0) {
                // Regular images
                GPlayerRelation rel = getPlayerRelation(s.getOwner());
                ctx.setColor(rel == is_Me ? COLOR_GREEN :
                             rel == is_Ally ? COLOR_YELLOW : COLOR_RED);
                for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
                    GfxPoint pt = scaleXY(chart_conf.getSimplePointAlias(s.getPos(), img));
                    drawShipIcon(ctx, rel, big, pt);
                }

                // Special case for circular wrao
                GPoint pt;
                if (chart_conf.getMode() == GChartConfiguration::Circular && chart_conf.getPointAlias(s.getPos(), pt, 1, true)) {
                    drawShipIcon(ctx, rel, big, scaleXY(pt));
                }
            }
        }
    }
    if (show & GChartOptions::co_Trails) {
        for (int i = 1; i <= NUM_SHIPS; ++i) {
            const GShip& sh = univ.getShip(i);
            for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
                drawShipVector(can, sh, univ.getTurnNumber(), img);
            }
        }
    } else if (vec_id > 0 && univ.isValidShipId(vec_id)) {
        const GShip& sh = univ.getShip(vec_id);
        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            drawShipVector(can, sh, univ.getTurnNumber(), img);
        }
    } else {
        /* no vectors */
    }
}

/** Draw ship marker for part II. Draws the dot, fleet leader mark and label. */
void
GChartViewport::drawShipMarker(GfxContext& ctx, const GShip& sh, GfxPoint pt, const int show)
{
    GfxCanvas& can = ctx.getCanvas();
    if ((show & GChartOptions::co_ShipDots) != 0 && sh.getOrbitPlanetId() == 0) {
        can.drawPixel(pt, ctx.getRawColor(), GFX_ALPHA_OPAQUE);
    }
    if (sh.isFleetLeader()) {
        can.drawPixel(pt.movedBy(-1, -1), ctx.getRawColor(), GFX_ALPHA_OPAQUE);
        can.drawPixel(pt.movedBy(+1, -1), ctx.getRawColor(), GFX_ALPHA_OPAQUE);
        can.drawPixel(pt.movedBy(-1, +1), ctx.getRawColor(), GFX_ALPHA_OPAQUE);
        can.drawPixel(pt.movedBy(+1, +1), ctx.getRawColor(), GFX_ALPHA_OPAQUE);
    }
    if ((show & GChartOptions::co_Labels) != 0) {
        if (sh.getOrbitPlanetId() != 0)
            pt.y += 4;
        if ((show & GChartOptions::co_ShipDots) == 0)
            pt.y += 1;
        outText(ctx, pt.x, pt.y, getLabel(ShipLabel, sh.getId()));
    }
}

/** Draw Ships, part II. These are drawn above the planets, so that planet
    rings do not obscure ships in warp wells, like in VPA, for example
    (hint hint! ;-) */
void
GChartViewport::drawShips(GfxCanvas& can, GUniverse& univ, int show)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setTextAlign(1, 0);
    ctx.useFont(*fonts[getFont()]);

    for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
        const GShip& s = univ.getShip(i);
        ctx.setColor(BLACK_COLOR_SET[getPlayerColor(s.getOwner())]);
        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            GfxPoint pt = scaleXY(chart_conf.getSimplePointAlias(s.getPos(), img));
            drawShipMarker(ctx, s, pt, show);
        }

        if (chart_conf.getMode() == GChartConfiguration::Circular) {
            GPoint pt;
            if (chart_conf.getPointAlias(s.getPos(), pt, 1, true)) {
                drawShipMarker(ctx, s, scaleXY(pt), show);
            }
        }
    }
}

/** Draw Warp Well.
    \param can Canvas to draw on
    \param pt Screen coordinates */
void
GChartViewport::drawWarpWell(GfxCanvas& can, GfxPoint pt)
{
    const int range = config.GravityWellRange();
    const int half = scale(1) / 2;

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_SHIELD+4);

    if (config.AllowGravityWells() && range > 0) {
        if (config.RoundGravityWells()) {
            /* This draws an edgy frame around the planet's warp well.
               At large zoom levels, each position in space becomes a
               "square" with an actual area; we draw the outline for
               these squares. For simplicity, the algorithm is split
               into 8 octants. */
            int wwx = range;
            int wwy = 0;

            int sx = scale(wwx) + half;
            int sy = scale(wwy) + half;
            drawVLine(ctx, pt.x - sx, pt.y - sy, pt.y + sy);
            drawVLine(ctx, pt.x + sx, pt.y - sy, pt.y + sy);
            drawHLine(ctx, pt.x - sy, pt.y - sx, pt.x + sy);
            drawHLine(ctx, pt.x - sy, pt.y + sx, pt.x + sy);

            while (wwx > wwy) {
                ++wwy;
                int nsy = scale(wwy) + half;
                if (square(wwx) + square(wwy) > square(range)) {
                    --wwx;
                    int nsx = scale(wwx) + half;
                    drawHLine(ctx, pt.x - sx, pt.y - sy, pt.x - nsx);
                    drawHLine(ctx, pt.x - sx, pt.y + sy, pt.x - nsx);
                    drawHLine(ctx, pt.x + sx, pt.y - sy, pt.x + nsx);
                    drawHLine(ctx, pt.x + sx, pt.y + sy, pt.x + nsx);
                    drawVLine(ctx, pt.x - sy, pt.y + sx, pt.y + nsx);
                    drawVLine(ctx, pt.x - sy, pt.y - sx, pt.y - nsx);
                    drawVLine(ctx, pt.x + sy, pt.y + sx, pt.y + nsx);
                    drawVLine(ctx, pt.x + sy, pt.y - sx, pt.y - nsx);
                    sx = nsx;
                }
                drawVLine(ctx, pt.x - sx, pt.y - sy, pt.y - nsy);
                drawVLine(ctx, pt.x - sx, pt.y + sy, pt.y + nsy);
                drawVLine(ctx, pt.x + sx, pt.y - sy, pt.y - nsy);
                drawVLine(ctx, pt.x + sx, pt.y + sy, pt.y + nsy);
                drawHLine(ctx, pt.x - sy, pt.y - sx, pt.x - nsy);
                drawHLine(ctx, pt.x - sy, pt.y + sx, pt.x - nsy);
                drawHLine(ctx, pt.x + sy, pt.y - sx, pt.x + nsy);
                drawHLine(ctx, pt.x + sy, pt.y + sx, pt.x + nsy);
                sy = nsy;
            }
        } else {
            /* Just a rectangle */
            const int dx = scale(range) + half;
            drawRectangle(ctx, GfxRect(pt.x - dx, pt.y - dx, 2*dx+1, 2*dx+1));
        }
    }
}

/** Draw planet marker. */
void
GChartViewport::drawPlanetMarker(GfxContext& ctx, const GPlanet& pl, GfxPoint pt, const int show)
{
    GfxCanvas& can = ctx.getCanvas();

    enum RingType { r_Enemy, r_Own, r_Unowned, r_Allied, r_Unknown };

    bool     base;          // true iff planet has base
    RingType ring;          // type of planet ring
    int      incy;          // Y increment for text
    bool     shipg;         // true iff ship ring is dotted (guessed)
    int      shipring;      // color of ship ring, 0 if none

    /* Compute decoration -- Planet */
    // FIXME: history & stuff
    if (pl.getOwner() == getPlayerId()) {
        ring = r_Own;
    } else {
        if (pl.getOwner())
            ring = (getPlayerRelation(pl.getOwner()) == is_Ally ? r_Allied : r_Enemy);
        else if (pl.hasAnyPlanetData())
            ring = r_Unowned;
        else
            ring = r_Unknown;
    }
    base = pl.hasBase();

    /* Compute decoration -- Ship */
    shipg = false;
    shipring = 0;
    switch(pl.getOrbitFlags() & (GPlanet::FlagOwnShipsInOrbit + GPlanet::FlagEnemyShipsInOrbit + GPlanet::FlagAlliedShipsInOrbit)) {
     case 0:
        shipg = true;
        switch(pl.getOrbitFlags() & (GPlanet::FlagGuessedEnemyInOrbit + GPlanet::FlagGuessedAllyInOrbit)) {
         case 0:
            shipring = 0;
            break;
         case GPlanet::FlagGuessedAllyInOrbit:
            shipring = BLACK_COLOR_SET[UIColor::tc_Yellow];
            break;
         case GPlanet::FlagGuessedEnemyInOrbit:
            shipring = COLOR_RED;
            break;
         default:
            shipring = COLOR_YELLOW;
            break;
        }
        break;
     case GPlanet::FlagOwnShipsInOrbit:
        shipring = COLOR_GREEN;
        break;
     case GPlanet::FlagEnemyShipsInOrbit:
        shipring = COLOR_RED;
        break;
     case GPlanet::FlagAlliedShipsInOrbit:
     case GPlanet::FlagAlliedShipsInOrbit + GPlanet::FlagOwnShipsInOrbit:
        shipring = BLACK_COLOR_SET[UIColor::tc_Yellow];
        break;
     default:
        shipring = COLOR_YELLOW;
        break;
    }

    incy = 1;

    /* Selection */
    if ((show & GChartOptions::co_Selection) && (pl.isMarked())) {
        ctx.setColor(COLOR_YELLOW);
        drawSelection(ctx, pt, mult, divi);
    }

    /* Warp Wells */
    if (show & GChartOptions::co_WarpWells)
        drawWarpWell(can, pt);

    /* Draw the planet */
    if (ring == r_Allied)
        ctx.setColor(COLOR_DARKYELLOWSCALE+7);
    else if (ring != r_Own)
        ctx.setColor(COLOR_BLUEBLACK);
    else
        ctx.setColor(COLOR_BLUE);
    if (base) {
        int sbsize = std::min(int(scale(SCCrossRadius)), int(SCMaxCrossRadius));
        drawHLine(ctx, pt.x - sbsize, pt.y, pt.x + sbsize);
        drawVLine(ctx, pt.x, pt.y - sbsize, pt.y + sbsize);
    }
    can.drawPixel(pt, standard_colors[COLOR_WHITE], GFX_ALPHA_OPAQUE);

    int r = std::min(int(scale(SCRingRadius)), int(SCMaxRingRadius));
    if (shipring == 0 && r == 0)
        r = 1;
    if (r) {
        if (ring == r_Enemy || ring == r_Own || ring == r_Allied) {
            drawCircle(ctx, pt, r);
            incy += r;
        } else if (ring == r_Unowned) {
            drawDottedCircle(ctx, pt, r);
            incy += r;
        }
    }
    if (shipring != 0) {
        ctx.setColor(shipring);
        ++r;
        if (shipg)
            drawDottedCircle(ctx, pt, r);
        else
            drawCircle(ctx, pt, r);
        incy = r+1;
    }

    ctx.setColor(COLOR_GRAY);
    if ((show & GChartOptions::co_Labels) != 0) {
        outText(ctx, pt.x, pt.y + incy, getLabel(PlanetLabel, pl.getId()));
    }
}

/** Draw Planets. */
void
GChartViewport::drawPlanets(GfxCanvas& can, GUniverse& univ, int show)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setTextAlign(1, 0);
    ctx.useFont(*fonts[getFont()]);
    for (int i = univ.ty_any_planets.findNextIndex(0); i != 0; i = univ.ty_any_planets.findNextIndex(i)) {
        GPlanet& p = univ.getPlanet(i);
        for (int img = getFirstImage(); img >= 0; img = getNextImage(img)) {
            GfxPoint pt = scaleXY(chart_conf.getSimplePointAlias(p.getPos(), img));
            drawPlanetMarker(ctx, p, pt, show);
        }

        GPoint pt;
        if (chart_conf.getMode() == GChartConfiguration::Circular && chart_conf.getPointAlias(p.getPos(), pt, 1, true)) {
            drawPlanetMarker(ctx, p, scaleXY(pt), show);
        }
    }
}

int
GChartViewport::getFont() const
{
    // IF view.mult > view.divi THEN BEGIN
    //   IF view.mult > 2*view.divi THEN BEGIN
    //     TextStyle := 0;
    //     AAColor := 0;
    //   END ELSE TextStyle := 3;
    // END ELSE BEGIN
    //   IF 2*view.mult < view.divi THEN BEGIN
    //     IF 3*view.mult < view.divi THEN opt.show := opt.show AND NOT co_Labels;
    //     TextStyle := 4
    //   END ELSE BEGIN
    //     TextStyle := 2;
    //   END;
    // END;

    if (mult > divi) {
        if (mult > 2*divi) {
            /* more then 2:1: 22 pt */
            return FONT_TITLE;
        } else {
            /* 1:1 up to 2:1: 16 pt */
            return FONT_NORMAL;
        }
    } else {
        if (2*mult < divi) {
            /* smaller than 1:2 */
            // FIXME: should be 6 or 8 pt font, as in PCC 1.x */
            return FONT_SMALL;
        } else {
            /* 1:2 up to 1:1: 12 pt */
            return FONT_SMALL;
        }
    }
}

/** Zoom in.
    \param bbox applicable bounding box (limits available zoom factors)
    \returns true iff zoom factor changed */
bool
GChartViewport::zoomIn(GChartBBox& bbox)
{
    if (divi > 1)
        --divi;
    else if (mult < bbox.getMaxZoom() * divi)
        ++mult;
    else
        return false;           // can't zoom in further

    /* Zoom range may be exceeded. For example, old zoom 5:2, limit 4;
       new zoom 5:1 */
    if (mult > divi * bbox.getMaxZoom()) {
        mult = bbox.getMaxZoom();
        divi = 1;
    }
    return true;
}

/** Zoom out.
    \param bbox applicable bounding box
    \returns true iff zoom factor changed */
bool
GChartViewport::zoomOut(GChartBBox& bbox)
{
    // This does not use the bounding box's min_zoom limit;
    // Instead, the transition regular->reduced chart is done by the starchart itself.
    // This way, we can zoom out navigation charts as far as we need.
    (void) bbox;
    if (mult > 1) {
        --mult;
        return true;
    } else if (divi < GChartBBox::MAX_ZOOM) {
        ++divi;
        return true;
    } else {
        return false;
    }
}

Signal0 GChartOptions::sig_changed;

static GChartOptions chart_options[3] = {
    { GChartOptions::co_Default, GChartOptions::co_Default & GChartOptions::co_Tristate },
    { GChartOptions::co_Default, GChartOptions::co_Default & GChartOptions::co_Tristate },
    { GChartOptions::co_Default, GChartOptions::co_Default & GChartOptions::co_Tristate }
};

static ValueBitsetParser  parse_chartopts  ("ion,mine,ufos,sectors,borders,drawings,selection,labels,trails,shipdots,warpwells,messages");
static ConfigIntRefOption opt_small_show   (getUserPreferences(), "Chart.Small.Show",   parse_chartopts, chart_options[0].show);
static ConfigIntRefOption opt_small_fill   (getUserPreferences(), "Chart.Small.Fill",   parse_chartopts, chart_options[0].fill);
static ConfigIntRefOption opt_normal_show  (getUserPreferences(), "Chart.Normal.Show",  parse_chartopts, chart_options[1].show);
static ConfigIntRefOption opt_normal_fill  (getUserPreferences(), "Chart.Normal.Fill",  parse_chartopts, chart_options[1].fill);
static ConfigIntRefOption opt_scanner_show (getUserPreferences(), "Chart.Scanner.Show", parse_chartopts, chart_options[2].show);
static ConfigIntRefOption opt_scanner_fill (getUserPreferences(), "Chart.Scanner.Fill", parse_chartopts, chart_options[2].fill);

/** Get starchart options for a particular chart widget.
    \param is_scanner  true if scanner in a control screen, false if main chart
    \param mult,divi   zoom settings

    Currently implements the PCC 1.x policy. */
GChartOptions&
getChartOpts(bool is_scanner, int mult, int divi)
{
    if (is_scanner)
        return chart_options[ScannerChartOptions];
    else if (mult < divi)
        return chart_options[SmallChartOptions];
    else
        return chart_options[NormalChartOptions];
}

GChartOptions&
getChartOpts(GChartOptionIndex index)
{
    return chart_options[index];
}

/** Toggle an option (or set of options).
    \param which_ones options to toggle, set of co_XXX */
void
GChartOptions::toggleOptions(int32 which_ones)
{
    // First, tristate options. We have three states:
    //    show   fill
    //      0      ?
    //      1      0
    //      1      1
    int32 tristate_opts   = which_ones & co_Tristate;
    int32 shown_tristate  = tristate_opts & show;        /* shown filled or shown empty */
    int32 hidden_tristate = tristate_opts & ~show;       /* hidden */
    int32 empty_tristate  = shown_tristate & ~fill;      /* shown filled */
    int32 filled_tristate = shown_tristate & fill;       /* shown empty */

    show |= hidden_tristate;                             /* show hidden */
    fill &= ~hidden_tristate;

    fill |= empty_tristate;                              /* fill empty */

    show &= ~filled_tristate;                            /* hide filled */
    fill &= ~filled_tristate;

    // Then, normal options
    int32 boolean_opts = which_ones & ~co_Tristate;
    show ^= boolean_opts;

    sig_changed.raise();
}

/** Toggle an option, given a key.
    \param c Key (modifiers removed)
    \retval true Key handled
    \retval false Key not recognized */
bool
GChartOptions::toggleOptionKey(uint32_t c)
{
    switch (c) {
     case 'm':
        toggleOptions(co_Mine);
        return true;
     case 'a':
        toggleOptions(co_ShipDots);
        return true;
     case 'd':
        toggleOptions(co_Labels);
        return true;
     case 'i':
        toggleOptions(co_Ion);
        return true;
     case 'v':
        toggleOptions(co_Trails);
        return true;
     case 't':
        toggleOptions(co_Selection);
        return true;
     case 's':
        toggleOptions(co_Sectors);
        return true;
     case 'b':
        toggleOptions(co_Borders);
        return true;
     case 'u':
        toggleOptions(co_Ufo);
        return true;
     case 'p':
        toggleOptions(co_Drawings);
        return true;
     case 'w':
        toggleOptions(co_WarpWells);
        return true;
     case 'n':
        toggleOptions(co_Messages);
        return true;
     default:
        return false;
    }
}

/** Copy from another option set. */
void
GChartOptions::copyFrom(const GChartOptions& opts)
{
    if (fill != opts.fill || show != opts.show) {
        fill = opts.fill;
        show = opts.show;
        sig_changed.raise();
    }
}

/** Copy values from another option set.
    \param opts Option set to copy from
    \param mask Mask describing options to copy */
void
GChartOptions::copyValues(const GChartOptions& opts, int32_t mask)
{
    const int32_t newShow =  (opts.show & mask) | (show & ~mask);
    const int32_t newFill = ((opts.fill & mask) | (fill & ~mask)) & co_Tristate;
    if (show != newShow || fill != newFill) {
        show = newShow;
        fill = newFill;
        sig_changed.raise();
    }
}

/** Set option. Set all options in \c mask to "enabled, not filled".
    This is a state all options allow.
    \param mask Options to set */
void
GChartOptions::setOption(int32_t mask)
{
    const int32_t newShow = (show | mask);
    const int32_t newFill = (fill & ~mask);
    if (show != newShow || fill != newFill) {
        show = newShow;
        fill = newFill;
        sig_changed.raise();
    }
}

static ConfigOption*const chartopt_ptrs[] = {
    &opt_small_show,
    &opt_small_fill,
    &opt_normal_show,
    &opt_normal_fill,
    &opt_scanner_show,
    &opt_scanner_fill,
};

int
getNumChartOptions()
{
    return countof(chartopt_ptrs);
}

ConfigOption&
getChartOption(int index)
{
    return *chartopt_ptrs[index];
}
