/**
  *  \file client/dialogs/consoledlg.cc
  *  \brief Console Dialog
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

/* Define this to get a line listing consumed CPU and real time after each statement */
#undef PROFILING

#ifdef PROFILING
# include <sys/time.h>
# include <time.h>
#endif

#include <climits>
#include <cctype>
#include <set>
#include "client/dialogs/consoledlg.h"
#include "client/dialogbinding.h"
#include "client/dialogs/helpdlg.h"
#include "client/widgets/consoleview.h"
#include "client/widgets/propertylist.h"
#include "cpluslib/format.h"
#include "cpluslib/utf8.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "int/bytecode.h"
#include "int/exec.h"
#include "int/expr.h"
#include "int/if/globalif.h"
#include "int/if/userif.h"
#include "int/interror.h"
#include "int/intglobal.h"
#include "int/keywords.h"
#include "int/lex.h"
#include "int/process.h"
#include "int/propenum.h"
#include "int/special.h"
#include "int/statement.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/inputline.h"
#include "ui/window.h"
#include "util/console.h"
#include "util/nls.h"
#include "util/strlist.h"
#include "ui/widgets/quit.h"

namespace {
    class WConsoleDialog : public UIWindow, public IntUserInterfaceDialogBinding {
        ObjectHolder h;
        UIInputLine* input;
        WConsoleView* cv;
        int recall_index;
        string_t recall_last_line;

        /** List of contexts. This one is used for completion,
            not for process creation. */
        ptr_vector<IntContext> contextCache;

        bool first;

        void init();
     public:
        WConsoleDialog();
        void onEnter();
        void doRecall(int direction);
        void doInsert();
        void doListVariables();
        bool handleEvent(const UIEvent& e, bool second_pass);

        // IntUserInterfaceDialogBinding
        virtual bool processPreIdle();
        virtual bool processEvent(std::auto_ptr<IntUserInterfaceEvent> event);
    };
}

/** "Completion" operation.
    \param input Input line
    \param contexts Active (non-global) contexts
    \param onlyCommands Expect only commands, no variables in the initial position */
void
doComplete(UIInputLine& input, const ptr_vector<IntContext>& contexts, bool onlyCommands)
{
    // Find word to complete. The string is encoded in UTF-8, and pos is a character number.
    // Use a simple forward algorithm instead of trying to parse UTF-8 backwards.
    // Note that this function assumes that all possible completions are ASCII.
    string_t line = input.getText();
    string_t::size_type pos = input.getCursor();
    string_t stem;
    bool acceptCommands = true;
    Utf8Reader rdr(line);

    enum {
        Normal,
        SeenConfigCommand,
        SeenConfigFunction,
        SeenConfigFunctionParen,
        SeenQuote
    } state = Normal;

    while (rdr.hasMore() && pos > 0) {
        // Letters, '$' and '_' can start a word, '0'..'9' and '.' can continue
        uint32_t ch = rdr.get();
        if ((ch >= 'A' && ch <= 'Z')
            || (ch >= 'a' && ch <= 'z')
            || ch == '_'
            || ch == '$'
            || (!stem.empty()
                && ((ch >= '0' && ch <= '9')
                    || ch == '.')))
        {
            // Valid word character
            stem += char(std::toupper(ch));
        } else {
            // Not a word.
            // Process previous word.
            if (!stem.empty()) {
                if (acceptCommands && stem == "ADDCONFIG") {
                    state = SeenConfigCommand;
                } else if (stem == "CFG") {
                    state = SeenConfigFunction;
                } else {
                    state = Normal;
                }
            }

            // Process non-word
            switch (ch) {
             case ' ':
                break;
             case '(':
                state = (state == SeenConfigFunction ? SeenConfigFunctionParen : Normal);
                break;
             case '"':
             case '\'':
                if (state == SeenConfigFunctionParen || state == SeenConfigCommand) {
                    state = SeenQuote;
                } else {
                    state = Normal;
                }
                break;
             default:
                state = Normal;
            }

            if (ch != ' ' || !stem.empty()) {
                acceptCommands = false;
            }
            stem.clear();
        }
        --pos;
    }

    // Do we have a stem? No completion of empty word.
    if (stem.empty()) {
        return;
    }

    // Gather completions
    struct Completer : public IntPropertyAcceptor {
        std::set<string_t> values;
        const string_t& stem;
        bool acceptCommands;
        bool onlyCommands;
        Completer(const string_t& stem, bool acceptCommands, bool onlyCommands)
            : stem(stem), acceptCommands(acceptCommands), onlyCommands(onlyCommands)
            { }
        void addProperty(const string_t& word, TIntTypeHint th)
            {
                // Only accept commands when valid at current place
                if (th == thProcedure && !acceptCommands) {
                    return;
                }

                // Do not accept non-commands when required
                if ((th != thProcedure && th != thNone) && acceptCommands && onlyCommands) {
                    return;
                }

                // Only accept correct stem
                if (word.size() < stem.size() || word.compare(0, stem.size(), stem, 0, stem.size()) != 0) {
                    return;
                }

                // Never offer as completion anything with an embedded (!= final) '$'.
                // This is used to avoid completing 'CC$foo' commands.
                string_t::size_type n = word.find('$', stem.size());
                if (n != word.npos && n != word.size()-1) {
                    return;
                }

                values.insert(beautifyName(word));
            }

        void addConfigOption(const string_t& word)
            {
                if (strUCase(word.substr(0, stem.size())) == stem) {
                    values.insert(word);
                }
            }
    };

    // Iterate through all contexts for completions. Note that IntGlobalContext
    // does not appear here, so it must be handled separately.
    Completer cpl(stem, acceptCommands, onlyCommands);
    if (state == SeenQuote) {
        for (Config::iterator it = config.begin(), e = config.end(); it != e; ++it) {
            cpl.addConfigOption(it->getName());
        }
    } else {
        for (uint32_t i = 0; i < contexts.size(); ++i) {
            contexts[i]->enumProperties(cpl);
        }
        IntGlobalContext().enumProperties(cpl);
        enumSpecialCommands(cpl);
        enumKeywords(cpl);
    }

    // Any completions?
    if (cpl.values.empty()) {
        return;
    }

    // Find common completion
    std::set<string_t>::iterator p = cpl.values.begin();
    const string_t& first = *p;
    string_t::size_type maxcomm = first.size();
    while (p != cpl.values.end()) {
        const string_t& me = *p;
        if (me.size() < maxcomm) {
            maxcomm = me.size();
        }
        for (string_t::size_type i = 0; i < maxcomm; ++i) {
            if (me[i] != first[i]) {
                maxcomm = i;
                break;
            }
        }
        ++p;
    }

    // Check common completion
    if (maxcomm > stem.size()) {
        // All completions have a common initial part
        input.insert(string_t(first, stem.size(), maxcomm - stem.size()));
    } else {
        // Different completions. Ask user.
        StringList slist;
        int key = 0;
        for (std::set<string_t>::iterator i = cpl.values.begin(); i != cpl.values.end(); ++i) {
            slist.add(key, *i);
            ++key;
        }
        UIStandardListbox box(0,
                              slist.getCount() < 5 ? 5 : slist.getCount() > 15 ? 15 : slist.getCount(),
                              getTextWidth(*fonts[FONT_NORMAL], slist) + 20,
                              slist);
        bool result = box.doStandardListbox(_("Completions"), "pcc2:console");

        // Must manually reset TypeErase which gets set by the focus change
        input.setState(UIInputLine::ilf_TypeErase, false);
        if (result) {
            input.insert(string_t(slist.getItem(box.getCurrent())->second, stem.size()));
        }
    }
}

/***************************** WConsoleDialog ****************************/

WConsoleDialog::WConsoleDialog()
    : UIWindow(_("Console"), UIWindow::BLUE_WINDOW, COLOR_GRAY), recall_index(-1),
      contextCache(),
      first(true)
{
    init();
}

void
WConsoleDialog::init()
{
    // VBox
    //   WConsoleViewWithHandler
    //   UIInputLine
    //   HBox
    //     UIButton "Evaluate"
    //     UIButton "Close"
    //     UISpacer
    //     UIButton "Help"
    setLayout(UIVBoxLayout::instance5);

    UIGroup& buttons = h.add(new UIGroup(UIHBoxLayout::instance5));

    cv = new WConsoleView(console, 10);
    add(h.add(cv));
    input = new UIInputLine(INT_MAX, 30, 0, UIInputLine::ilf_Framed, 0);
    add(h.add(input));
    add(buttons);
    add(h.add(new UIQuit(1)));

    UIButton& button_eval  = h.add(new UIButton(1, _("Evaluate"), SDLK_RETURN));
    UIButton& button_close = h.add(new UIButton(1, _("Close"), SDLK_ESCAPE));
    buttons.add(button_eval);
    buttons.add(button_close);
    buttons.add(h.add(new UISpacer()));
    buttons.add(h.add(new UIButton(_("Help"), ss_Alt + 'h')));

    button_close.addStop();
    button_eval.addCall(this, &WConsoleDialog::onEnter);

    input->setFocusedWidget();

    pack();

    enumContexts(contextCache);

    // Enable synchronous mode. There's no reason not to do this.
    // Originally, synchronous mode was enabled during execution of commands.
    // However, since almost nothing is written to the console while no
    // command is executing, having it on all the time doesn't make a big
    // difference, but saves a lot of hassle.
    cv->setSynchronousMode(true);
}

/** Evaluate entered expression. */
void
WConsoleDialog::onEnter()
{
    cv->scrollToEnd();

    // Reset recall
    recall_index = -1;

    // Compile
    try {
        // Make bytecode object
        IntBCORef bco = new IntBytecodeObject();

        // Make execution context containing all current contexts
        IntExecutionContext& exec = createProcess(format(_("Console: %s"), input->getText()));
        ptr_vector<IntContext> ctx;
        enumContexts(ctx);
        exec.pushContextsFrom(ctx);
        exec.markContextTOS();

        // Log command if it appears nonempty
        string_t cmd = input->getText();
        string_t::size_type first = cmd.find_first_not_of(" ");
        if (first != string_t::npos && cmd[first] != '%') {
            console.write(LOG_INFO, cmd, ConsoleRecallable);
        }

        // Make command source containing this text
        IntMemoryCommandSource mcs(cmd);

        // Make compilation context; configure for one-liners
        IntStatementCompilationContext scc;
        scc.withExecutionContext(&exec);
        scc.withFlags(IntCompilationContext::RefuseBlocks);
        scc.withFlags(IntCompilationContext::LinearExecution);

        // Compile
        IntStatementCompiler sc(mcs);
        IntStatementCompiler::StatementResult result = sc.compile(*bco, scc);

        if (bco->getNumInstructions() == 0) {
            // blank input
        } else {
            // valid input, execute
            sc.finishBCO(*bco);
            exec.pushNewFrame(new IntExecutionFrame(bco));
#ifdef PROFILING
            clock_t beg = clock();
            struct timeval beg_tv;
            gettimeofday(&beg_tv, 0);

            runRunnableProcesses();

            clock_t end = clock();
            struct timeval end_tv;
            gettimeofday(&end_tv, 0);

            if (end_tv.tv_usec < beg_tv.tv_usec) {
                end_tv.tv_usec += 1000000;
                end_tv.tv_sec--;
            }
            int elapsedCPU = int(1000*double(end - beg)/CLOCKS_PER_SEC);
            int elapsedWCT = 1000*(end_tv.tv_sec - beg_tv.tv_sec) + (end_tv.tv_usec - beg_tv.tv_usec) / 1000;

            console.write(LOG_DEBUG, format("Evaluation took %d ms CPU, %d ms real", elapsedCPU, elapsedWCT));
#else
            runRunnableProcesses();
#endif
            switch (exec.getState()) {
             case IntExecutionContext::psSuspended:
                // Suspended
                console.write(LOG_INFO, _("Suspended."), UIColor::tc_Faded);
                break;
             case IntExecutionContext::psError:
                // Hand this to our normal error handler
                throw exec.getError();
             case IntExecutionContext::psEnded:
                // Ended regularily
                if (result == IntStatementCompiler::CompiledExpression) {
                    // expression, print result
                    IntValue* result = exec.getResult();
                    if (result != 0)
                        console.write(LOG_INFO, result->toString(true), ConsoleRightJustified | ConsoleRecallable);
                    else
                        console.write(LOG_INFO, "Empty", UIColor::tc_Faded | ConsoleRightJustified);
                }
                break;
             case IntExecutionContext::psTerminated:
                // Terminated, i.e. "End" statement. Log only when user specified an expression,
                // to tell them why they don't get a result.
                if (result == IntStatementCompiler::CompiledExpression) {
                    console.write(LOG_INFO, _("Terminated."), UIColor::tc_Faded);
                }
                break;
             default:
                // Impossible or uninteresting state
                break;
            }
            killTerminatedProcesses();
        }

        // Update screen
        if (haveDisplayedTurn()) {
            getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
        }

        // Clear input line for next expression
        input->setText("");

        // Trigger continuation of scripts
        triggerScripts();
    }
    catch (IntError& e) {
        console.write(LOG_ERROR, e.what(), ConsoleRightJustified);

        string_t trace = e.getTrace();
        if (trace.size()) {
            console.write(LOG_INFO, trace, ConsoleRightJustified);
        }
        killTerminatedProcesses();
    }
    catch (std::exception& e) {
        console.write(LOG_ERROR, e.what());
        killTerminatedProcesses();
    }
}

/** "Recall operation": browse through history of previous recallable input and output. */
void
WConsoleDialog::doRecall(int direction)
{
    int index = recall_index;
    if (index < 0)
        /* Remember final input */
        recall_last_line = input->getText();
    while (1) {
        index += direction;
        if (index >= console.getScrollbackFill()) {
            /* We cannot scroll so far back into the past */
            break;
        }
        if (index < 0) {
            /* We have reached the present */
            if (recall_index >= 0) {
                /* We were in the past */
                input->setText(recall_last_line);
            }
            recall_index = -1;
            break;
        }
        if (console.getLine(index).format & ConsoleRecallable) {
            /* We have found a message that can be recalled */
            recall_index = index;
            input->setText(console.getLine(index).text);
            break;
        }
    }
}

/** "Insert" operation: insert last result into current input. */
void
WConsoleDialog::doInsert()
{
    for (int index = 0; index < console.getScrollbackFill(); ++index) {
        if (console.getLine(index).format == (ConsoleRightJustified | ConsoleRecallable)) {
            /* This is our last result */
            input->setSelectedWidget();
            input->insert(console.getLine(index).text);
            break;
        }
    }
}

void
WConsoleDialog::doListVariables()
{
    IntVariableNames* names = 0;
    IntDataSegment* data = 0;
    const char* ttl = 0;

    // Locate the current object
    for (uint32_t i = contextCache.size(); i > 0; --i) {
        GObject* obj = contextCache[i-1]->getObject();
        if (dynamic_cast<GPlanet*>(obj)) {
            names = &planet_property_names;
            data = &getPlanetProperties(obj->getId());
            ttl = N_("Planet Properties");
            break;
        }
        if (dynamic_cast<GShip*>(obj)) {
            names = &ship_property_names;
            data = &getShipProperties(obj->getId());
            ttl = N_("Ship Properties");
            break;
        }
    }

    // Found anything?
    if (!names || names->getNumNames() == 0) {
        return;
    }

    // Dialog
    WPropertyList plist(0, *names, *data);
    bool result = plist.doStandardListbox(_(ttl), "pcc2:console");

    // Must manually reset TypeErase which gets set by the focus change
    input->setState(UIInputLine::ilf_TypeErase, false);
    if (result) {
        input->insert(beautifyName(names->lookupAddress(plist.getCurrent())));
    }
}

bool
WConsoleDialog::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case ss_Shift + SDLK_PAGEUP:
            cv->setSelectedWidget();
            cv->scrollBack(5);
            return true;
         case ss_Shift + SDLK_UP:
            cv->setSelectedWidget();
            cv->scrollBack(1);
            return true;
         case ss_Shift + SDLK_PAGEDOWN:
            cv->setSelectedWidget();
            cv->scrollBack(-5);
            return true;
         case ss_Shift + SDLK_DOWN:
            cv->setSelectedWidget();
            cv->scrollBack(-1);
            return true;
         case SDLK_UP:
            doRecall(+1);
            return true;
         case SDLK_DOWN:
            doRecall(-1);
            return true;
         case SDLK_INSERT:
            doInsert();
            return true;
         case SDLK_TAB:
            doComplete(*input, contextCache, false);
            return true;
         case ss_Alt + 'v':
            doListVariables();
            return true;
         case SDLK_F1:
         case ss_Alt + 'h':
            doHelp("pcc2:console");
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

bool
WConsoleDialog::processPreIdle()
{
    if (first) {
        const char* message = _("Enter command or expression:");
        if (console.getScrollbackFill() == 0 || console.getLine(0).text != message) {
            console.write(LOG_INFO, message);
        }
        first = false;
    }
    return false;
}

bool
WConsoleDialog::processEvent(std::auto_ptr<IntUserInterfaceEvent> event)
{
    switch (event->command) {
     case iucPopupConsole:
        // Nothing to do, we're already here
        return true;

     case iucExitRace:
     case iucExitClient:
     case iucOverlayMessage:
     case iucChartSetView:
     case iucGoToScreen:
     case iucGoToChart:
     case iucGoToMinefield:
     case iucGoToUfo:
     case iucGoToStorm:
     case iucUseKeymap:
     case iucExitDialog:
        // Default handling
        return IntUserInterfaceDialogBinding::processEvent(event);
    }
    return false;
}

/** Console, main entry. */
void
doConsole()
{
    WConsoleDialog win;
    ui_root->centerWidget(win);
    win.runModal(win);
}
