/**
  *  \file client/playerscr.cc
  *  \brief Player Main Screen
  *
  *  Main screen, user interface and multiplexing to other control screens.
  *  PCC 1.x source: cc.pas, scrmux.pas.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/playerscr.h"
#include "client/chartusr.h"
#include "client/dialogbinding.h"
#include "client/dialogs/combatdlg.h"
#include "client/dialogs/commandedit.h"
#include "client/dialogs/consoledlg.h"
#include "client/dialogs/gadialog.h"
#include "client/dialogs/histselect.h"
#include "client/dialogs/imperial.h"
#include "client/dialogs/notifydlg.h"
#include "client/dialogs/prefdlg.h"
#include "client/dialogs/processmgr.h"
#include "client/dialogs/teamedit.h"
#include "client/scr-score.h"
#include "client/scr-sim.h"
#include "client/screens.h"
#include "client/screens/basescreen.h"
#include "client/screens/basetaskscreen.h"
#include "client/screens/fleetscreen.h"
#include "client/screens/histscreen.h"
#include "client/screens/planetscreen.h"
#include "client/screens/planettaskscreen.h"
#include "client/screens/shipscreen.h"
#include "client/screens/shiptaskscreen.h"
#include "client/vcrplay.h"
#include "client/widgets/decaymsg.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "cpluslib/ptrmultilist.h"
#include "game/classicvcr.h"
#include "game/fleet.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/racename.h"
#include "game/ship.h"
#include "game/sim-game.h"
#include "game/stat.h"
#include "game/team.h"
#include "gfx/font.h"
#include "gfx/gfxcompl.h"
#include "gfx/palette.h"
#include "int/if/ifutil.h"
#include "int/if/userif.h"
#include "int/interror.h"
#include "int/notify.h"
#include "resmgr/resmgr.h"
#include "ui/layout.h"
#include "ui/ui.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "util/nls.h"
#include "client/screens/minescreen.h"
#include "client/screens/ionscreen.h"
#include "client/screens/ufoscreen.h"
#include "int/simple.h"
#include "ui/uiglobal.h"
#ifdef CONF_FLAK_SUPPORT
# include "flak/dlg-select.h"
# include "flak/flak-db.h"
# include "flak/vis-ui.h"
#endif

class WPixmapButton : public UIButton {
    Ptr<GfxPixmap> pix;
 public:
    WPixmapButton(int aid, string_t aname, UIKey akey, Ptr<GfxPixmap> apix);
    void drawContent(GfxCanvas& can);
    void getLayoutInfo(LayoutInfo& info);
};

WPixmapButton::WPixmapButton(int aid, string_t aname,
                             UIKey akey, Ptr<GfxPixmap> apix)
    : UIButton(aid, aname, akey), pix(apix)
{ }

void
WPixmapButton::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useFont(*fonts[FONT_SMALL]);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    if (pix) {
        pix->blit(can, extent.x + 1, extent.y + 1, GfxRect(0, 0, extent.w - 2, extent.h - 2));
    }

    enum { N = 9 };
    static const uint8_t spec[N][4] = {
        { 0, 0, COLOR_BLACK, 128 },
        { 0, 2, COLOR_BLACK, 128 },
        { 2, 0, COLOR_BLACK, 128 },
        { 2, 2, COLOR_BLACK, 128 },
        { 0, 1, COLOR_BLACK, GFX_ALPHA_OPAQUE },
        { 1, 0, COLOR_BLACK, GFX_ALPHA_OPAQUE },
        { 2, 1, COLOR_BLACK, GFX_ALPHA_OPAQUE },
        { 1, 2, COLOR_BLACK, GFX_ALPHA_OPAQUE },
        { 1, 1, COLOR_WHITE, GFX_ALPHA_OPAQUE }
    };

    int x = extent.x;
    int y = extent.y + extent.h - ctx.getFont().getTextHeight(getName());
    for (uint_t i = 0; i < sizeof(spec)/sizeof(spec[0]); ++i) {
        ctx.setColor(spec[i][2]);
        ctx.setAlpha(spec[i][3]);
        outTextF(ctx, x + spec[i][0], y - spec[i][1], extent.w - 2, getName());
    }
    drawFrameDown(can, extent);
    can.update(extent);
}

void
WPixmapButton::getLayoutInfo(LayoutInfo& info)
{
    GfxPoint pt (fonts[FONT_SMALL]->getTextWidth(getName()) + 2,
                 fonts[FONT_SMALL]->getTextHeight(getName()) + 2);
    if (pix) {
        pt.x = pix->getWidth() + 2;
        pt.y = pix->getHeight();
    }
    info.flags = 0;
    info.min_size = info.pref_size = pt;
}

namespace {
    class WVcrSelectorWindowRealGameInterface : public WVcrSelectorWindowGameInterface {
     public:
        WVcrSelectorWindowRealGameInterface();

        bool isGameObject(const GVcrObject& obj);
        void goToGameObject(const GVcrObject& obj);

        void postEvent();

     private:
        GObject* target_obj;
    };
}

WVcrSelectorWindowRealGameInterface::WVcrSelectorWindowRealGameInterface()
    : target_obj(0)
{ }

bool
WVcrSelectorWindowRealGameInterface::isGameObject(const GVcrObject& obj)
{
    if (obj.isPlanet()) {
        return current_anyplanet_type.isValidIndex(obj.getId());
    } else {
        if (current_anyship_type.isValidIndex(obj.getId())) {
            GShip* sh = dynamic_cast<GShip*>(&current_anyship_type.getObjectByIndex(obj.getId()));
            return sh != 0
                && sh->getHullId().isKnown()
                && obj.canBeHull(sh->getHullId());
        } else {
            return false;
        }
    }
}

void
WVcrSelectorWindowRealGameInterface::goToGameObject(const GVcrObject& obj)
{
    if (obj.isPlanet()) {
        target_obj = &current_anyplanet_type.getObjectByIndex(obj.getId());
    } else {
        target_obj = &current_anyship_type.getObjectByIndex(obj.getId());
    }
}

void
WVcrSelectorWindowRealGameInterface::postEvent()
{
    if (target_obj) {
        postGoToObject(IntUserInterfaceBinding::get(), *target_obj, false);
    }
}

/***************************** WPlayerScreen *****************************/

class WPlayerScreen : public UIGroup, public IntUserInterfaceDialogBinding {
    Ptr<GfxPixmap> pix;
    ObjectHolder holder;
    IntKeymapRef keymap;
    bool first;
    bool exitClient;

    void init();
 public:
    WPlayerScreen();
    void drawContent(GfxCanvas& can);
    bool handleEvent(const UIEvent& e, bool second_pass);

    bool wantExitClient() const
        { return exitClient; }

    virtual bool processEvent(std::auto_ptr<IntUserInterfaceEvent> event);
    virtual bool processPreIdle();

    virtual IntValue* getProperty(IntUserInterfaceProperty p);
    virtual void setProperty(IntUserInterfaceProperty p, IntValue* v);
};

WPlayerScreen::WPlayerScreen()
    : UIGroup(UINullLayout::instance /* will be overridden */),
      keymap(IntKeymap::getKeymapByName("RACESCREEN")),
      first(true),
      exitClient(false)
{
    init();
}

void
WPlayerScreen::init()
{
    UIVBoxLayout& vlayout = holder.add(new UIVBoxLayout(5, 5));
    setLayout(holder.add(new UIHBoxLayout(5, 5)));
    setExtent(GfxRect(*ui_screen));
    pix = requestPixmap("bgtile");
    UIGroup& g = holder.add(new UIGroup(vlayout));
    g.add(holder.add(new WPixmapButton(1, _("F1 - Starships"), SDLK_F1,
                             requestPixmap("menu.ship"))).addKey());
    g.add(holder.add(new WPixmapButton(2, _("F2 - Planets"), SDLK_F2,
                             requestPixmap("menu.planet"))).addKey());
    g.add(holder.add(new WPixmapButton(3, _("F3 - Starbases"), SDLK_F3,
                             requestPixmap("menu.base"))).addKey());
    g.add(holder.add(new WPixmapButton(4, _("F4 - Starchart"), SDLK_F4,
                             requestPixmap("menu.chart"))).addKey());
    g.add(holder.add(new UISpacer()));
    add(g);
    g.pack();

    UIGroup& fgroup = holder.add(new UIGroup(holder.add(new UIFlowLayout(7, true))));
    fgroup.add(holder.add(new UIButton(_("A - Alliances"), 'a')));
    fgroup.add(holder.add(new UIButton(_("Ctrl-O - Options"), ss_Ctrl + 'o')));
    fgroup.add(holder.add(new UIButton(_("F7 - Search"), SDLK_F7)));
    fgroup.add(holder.add(new UIButton(_("G - Global"), 'g')));
    fgroup.add(holder.add(new UIButton(_("T - Teams"), 't')));
    fgroup.add(holder.add(new UIButton(_("S - Scores"), 's')));
    fgroup.add(holder.add(new UIButton(_("B - Battle Simulator"), 'b')));
    fgroup.add(holder.add(new UIButton(_("I - Imperial Stats"), 'i')));
    fgroup.add(holder.add(new UIButton(_("W - Write Message"), 'w')));
    fgroup.add(holder.add(new UIButton(_("M - Messages"), 'm')).setState(st_Disabled, getDisplayedTurn().getInbox().getCount() == 0));
    fgroup.add(holder.add(new UIButton(_("V - Combat Recorder"), 'v')).setState(st_Disabled, getDisplayedTurn().getVcrs().getNumBattles() == 0));
    fgroup.add(holder.add(new UIButton(_("ESC - Exit"), SDLK_ESCAPE)));
    fgroup.add(holder.add(new UIButton(_("H - Help"), 'h')));

    UIGroup& vgroup = holder.add(new UIGroup(vlayout));
    vgroup.add(holder.add(new UISpacer()));
    vgroup.add(fgroup);
    add(vgroup);

    doLayout();
}

void
WPlayerScreen::drawContent(GfxCanvas& can)
{
    drawTiledArea(can, GfxRect(extent.x + 1, extent.y + 1,
                               extent.w - 2, extent.h - 2),
                  pix, standard_colors[COLOR_DARK], 16);
    drawFrameUp(can, extent);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_WHITE);

    ctx.useFont(*fonts[FONT_TITLE_BOLD]);
    outText(ctx, 200,  20, player_racenames.getFullName(getPlayerId()));

    ctx.useFont(*fonts[FONT_TITLE]);
    outText(ctx, 200,  60, format(_("Turn %d"), getDisplayedTurn().getTurnNumber()));
    outText(ctx, 200,  80, format(_("%d message%!1{s%}"), getDisplayedTurn().getInbox().getCount()));
    outText(ctx, 200, 100, format(_("%d combat recording%!1{s%}"), getDisplayedTurn().getVcrs().getNumBattles()));

    can.update(extent);
}

#ifdef CONF_FLAK_SUPPORT
namespace {
    class FlakVcrSelectorWindow : public UIWindow {
        FlakVcrSelector& sel;
     public:
        FlakVcrSelectorWindow(FlakVcrSelector& sel)
            : UIWindow(_("Visual Combat Recorder"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
              sel(sel)
            { }
        bool handleEvent(const UIEvent& event, bool second_pass);
        void playCurrent();
    };
}

bool
FlakVcrSelectorWindow::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case '+':
         case SDLK_DOWN:
         case SDLK_PAGEDOWN:
         case vk_WheelDown:
            sel.next(event.key.arg);
            return true;

         case '-':
         case SDLK_UP:
         case SDLK_PAGEUP:
         case vk_WheelUp:
            sel.previous(event.key.arg);
            return true;

         case SDLK_RETURN:
            playCurrent();
            return true;

         case vk_Quit:
            stop(cm_Escape);
            putKey(vk_Quit);
            return true;

         case SDLK_HOME:
         case ss_Ctrl + SDLK_PAGEUP:
         case ss_Ctrl + SDLK_HOME:
            sel.setCurrent(0);
            return true;

         case SDLK_END:
         case ss_Ctrl + SDLK_PAGEDOWN:
         case ss_Ctrl + SDLK_END:
            sel.setCurrent(sel.getDatabase().getNumBattles() - 1);
            return true;
        }
    }
    return UIWindow::handleEvent(event, second_pass);
}

void
FlakVcrSelectorWindow::playCurrent()
{
    playFlakBattle(sel.getDatabase().getBattle(sel.getCurrent()));
}

static void
playFlak(GFlakVcrDatabase& db, int index)
{
    ObjectHolder h;

    FlakVcrSelector sel(db);
    FlakVcrSelectorWindow w(sel);

    UIGroup but(UIHBoxLayout::instance5);
    but.add(h.add(new UIButton("\xE2\x86\x91", '-')));
    but.add(h.add(new UIButton("\xE2\x86\x93", '+')));
    but.add(h.add(new UIButton(_("ENTER - Play"), SDLK_RETURN)));
    but.add(h.add(new UISpacer()));
    but.add(h.add(new UIButton(cm_Escape, _("ESC - Exit"), SDLK_ESCAPE)).addStop());

    if (index)
        sel.setCurrent(index - 1);

    w.setLayout(UIVBoxLayout::instance5);
    w.setState(UIWindow::st_Modal, false);
    w.add(sel);
    w.add(but);
    w.pack();

    ui_root->centerWidget(w);
    ui_root->add(w);
    if (index)
        w.playCurrent();
    ui_root->run();
}
#endif

/** Play VCRs. */
static void
playVcrs(GVcrDatabase& db)
{
    if (GClassicVcrDatabase* cdb = dynamic_cast<GClassicVcrDatabase*>(&db)) {
        ObjectHolder holder;
        WVcrSelector sel(*cdb);
        WVcrSelectorWindowRealGameInterface gameif;
        WVcrSelectorWindow w(sel, gameif);

        UIGroup but(UIHBoxLayout::instance5);
        but.add(holder.add(new UIButton("\xE2\x86\x91", '-')));
        but.add(holder.add(new UIButton("\xE2\x86\x93", '+')));
        but.add(holder.add(new UIButton(_("ENTER - Play"), SDLK_RETURN)));
        but.add(holder.add(new UIButton(_("ESC - Exit"), SDLK_ESCAPE)));
        but.add(holder.add(new UISpacer()));
        but.add(holder.add(new UIButton(_("Help"), 'h')));

        w.setLayout(UIVBoxLayout::instance5);
        w.add(sel);
        w.add(but);
        w.add(holder.add(new WHelpWidget("pcc2:vcr")));
        w.pack();
        ui_root->centerWidget(w);
        ui_root->add(w);
        int result = ui_root->run();
        ui_root->close(w);
        if (result == cm_Enter) {
            gameif.postEvent();
        }
#ifdef CONF_FLAK_SUPPORT
    } else if (GFlakVcrDatabase* fdb = dynamic_cast<GFlakVcrDatabase*>(&db)) {
        playFlak(*fdb, 0);
#endif
    }
}

bool
WPlayerScreen::processEvent(std::auto_ptr<IntUserInterfaceEvent> event)
{
    switch (event->command) {
     case iucPopupConsole:
        doConsole();
        return false;

     case iucExitRace:
        return true;

     case iucExitClient:
        exitClient = true;
        return true;

     case iucOverlayMessage:
        showDecayingMessage(IntValue::toStringOf(event->args[0], false));
        return false;

     case iucChartSetView:
        // Nothing to do, command refuses internally
        return false;

     case iucGoToScreen:
        {
            int32_t value = 0;
            checkIntArg(value, event->args[0]);
            switch (value) {
             case 1:  doShipScreen();           break;
             case 2:  doPlanetScreen();         break;
             case 3:  doBaseScreen();           break;
             case 4:  doStarchart();            break;
             case 6:  doHistoryScreen();        break;
             case 10: doFleetScreen();          break;
             case 11: doShipAutoTaskScreen();   break;
             case 12: doPlanetAutoTaskScreen(); break;
             case 13: doBaseAutoTaskScreen();   break;
            }
        }
        return false;

     case iucGoToChart:
        doStarchart();
        return false;

     case iucGoToMinefield:
        doMinefieldScreen();
        return false;

     case iucGoToUfo:
        doUfoScreen();
        return false;

     case iucGoToStorm:
        doIonStormScreen();
        return false;

     case iucExitDialog:
        return false;

     case iucUseKeymap:
        return IntUserInterfaceDialogBinding::processEvent(event);
    }
    return false;
}

bool
WPlayerScreen::processPreIdle()
{
    if (first) {
        first = false;
        if (IntNotificationMessageStore::instance.getNumMessages() != 0) {
            doNotifyScreen();
        }
        return true;
    } else {
        return false;
    }
}

IntValue*
WPlayerScreen::getProperty(IntUserInterfaceProperty p)
{
    switch (p) {
     case iuiScreenNumber:
     case iuiScreenRegistered:
        return makeIntValue(0);

     case iuiSimFlag:
        return makeBoolValue(false);

     case iuiChartX:
     case iuiScanX:
        return makeIntValue(getStarchartPosition().x);

     case iuiChartY:
     case iuiScanY:
        return makeIntValue(getStarchartPosition().y);

     case iuiKeymap:
        return makeKeymapValue(keymap);
    }
    return 0;
}

void
WPlayerScreen::setProperty(IntUserInterfaceProperty p, IntValue* v)
{
    int32_t n;
    switch (p) {
     case iuiChartX:
     case iuiScanX:
        if (checkIntArg(n, v, 0, 10000)) {
            setStarchartPosition(GPoint(n, getStarchartPosition().y));
        }
        break;

     case iuiChartY:
     case iuiScanY:
        if (checkIntArg(n, v, 0, 10000)) {
            setStarchartPosition(GPoint(getStarchartPosition().x, n));
        }
        break;

     default:
        throw IntError::notAssignable();
    }
}

bool
WPlayerScreen::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        if (checkKeymap(e.key.code, e.key.arg)) {
            return true;
        } else if (e.key.code == 'v') {
            if (getDisplayedTurn().getVcrs().getNumBattles())
                playVcrs(getDisplayedTurn().getVcrs());
            return true;
        } else if (e.key.code == 'b') {
            GSimulatorRealGameInterface rgi(getDisplayedTurn().getCurrentUniverse());
            postGoToScreen(*this, doSimulator(rgi), 0);
            return true;
        } else if (e.key.code == 's') {
            if (isStatisticsFileLoaded() && getStatistics().getNumTurns() != 0)
                doScoreScreen(getStatistics(), getPlayerId());
            return true;
        } else if (e.key.code == 't') {
            editTeams();
            return true;
        } else if (e.key.code == 'i') {
            doImperialStatistics();
            return true;
        } else if (e.key.code == 'g') {
            doGlobalActions();
            return true;
        } else if (e.key.code == ss_Alt + 'e') {
            editCommands();
            return true;
        } else if (e.key.code == ss_Alt + 'q') {
            doProcessManager(0);
            return true;
        } else if (e.key.code == ss_Alt + 'n') {
            doNotifyScreen();
            return true;
        }
    }
    return UIGroup::handleEvent(e, second_pass);
}

/***************************** Main Routines *****************************/

/** Player screen.
    \retval true Exit client
    \retval false Proceed with game selection */
bool
doPlayerScreen()
{
    WPlayerScreen ps;

    // Hooks must be run while the player screen is active, so that correct UI properties are seen.
    runHook("LOAD");
    if (getCurrentTurn().getDatabaseTurn() < getCurrentTurn().getTurnNumber())
        runHook("NEWTURN");

    ps.runModal(ps);

    runHook("EXIT");

    return ps.wantExitClient();
}
