/**
  *  \file client/tiles/fleetwaypoint.cc
  *  \brief Fleet Waypoint Tile
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "client/tiles/fleetwaypoint.h"
#include "util/strtab.h"
#include "ui/util.h"
#include "game/mission.h"
#include "util/nls.h"
#include "ui/uiglobal.h"
#include "game/ship.h"
#include "game/config.h"
#include "game/global.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "ui/widgets/button.h"
#include "game/shippredict.h"
#include "util/misc.h"

static const char*const labels[] = {
    N_("Waypoint:"),
    N_("Warp Factor:"),
    N_("Distance:"),
    N_("E.T.A.:"),
    N_("Fuel Usage:"),
};

static int
tileHeight()
{
    return 5*font_heights[FONT_NORMAL] + font_heights[FONT_TITLE] + 5;
}

WFleetWaypointTile::WFleetWaypointTile(GObjectSelection& selection)
    : WObjectSelectionObserverTile(GfxPoint(315, tileHeight()), selection)
{
    init();
}

void
WFleetWaypointTile::init()
{
    setLayout(RightJustified, 5, true);

    int y = tileHeight();

    addControl(GfxRect(0,  y-20, 20, 20), h.add(new UIButton("U", 'u')));
    addControl(GfxRect(25, y-20, 20, 20), h.add(new UIButton("C", 'c')));
    addControl(GfxRect(0,  y-42, 20, 20), h.add(new UIButton("D", 'd')));
    addControl(GfxRect(25, y-42, 20, 20), h.add(new UIButton("A", 'a')));
    addControl(GfxRect(25, y-64, 20, 20), h.add(new UIButton("W", 'w')));
}

void
WFleetWaypointTile::drawData(GfxCanvas& can)
{
    // Commands in 1.x:
    //           W-warp
    //           Q-menu
    //  D-distr  A-aim
    //  U-unload C-cargo

    // Objects. The tile sits on the fleet member
    GShip* member = dynamic_cast<GShip*>(getCurrentObject());
    if (!member || member->getFleetNumber() == 0) {
        return;
    }
    GShip* leader = &getCurrentUniverse()->getShip(member->getFleetNumber());

    // Layout
    StaticTranslatedStringTable tab(labels, countof(labels));
    const int lwidth = getTextWidth(*fonts[FONT_NORMAL], tab) + 10;

    // Prepare draw
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setSolidBackground();
    int y = 0;
    int x0 = 5;
    int x1 = lwidth+5;
    int w = getInnerSize(0, 0).w - lwidth - 10;
    int h = font_heights[FONT_NORMAL];

    // Line 1: waypoint
    ctx.setColor(UIColor::tc_Static);
    outTextF(ctx, x0, y, lwidth, tab(0));

    ctx.setColor(UIColor::tc_Green);
    const GMission* msn = GMissionList::getInstance().getMissionByShip(*leader);
    if (msn != 0 && (msn->getFlags() & GMission::mf_Waypoint) != 0) {
        outTextF(ctx, x1, y, w, msn->getLabel(*leader));
    } else {
        outTextF(ctx, x1, y, w, getDisplayedTurn().getCurrentUniverse().getLocationName(leader->getWaypoint(), locs_WW | locs_Verbose));
    }
    y += h;

    // Line 2: speed
    ctx.setColor(UIColor::tc_Static);
    outTextF(ctx, x0, y, lwidth, tab(1));

    ctx.setColor(UIColor::tc_Green);
    if (leader->getWarp() == 0) {
        outTextF(ctx, x1, y, w, _("not moving"));
    } else {
        outTextF(ctx, x1, y, w, format(_("Warp %d"), leader->getWarp()));
    }
    y += h;
    y += 5;

    // Line 3: title
    ctx.useFont(*fonts[FONT_TITLE]);
    ctx.setColor(UIColor::tc_White);
    outTextF(ctx, x0, y, w+lwidth, _("This ship:"));
    y += font_heights[FONT_TITLE];
    ctx.useFont(*fonts[FONT_NORMAL]);

    // Line 4: distance
    ctx.setColor(UIColor::tc_Static);
    outTextF(ctx, x0, y, lwidth, tab(2));

    ctx.setColor(UIColor::tc_Green);
    outTextF(ctx, x1, y, w, format(_("%.2f ly"), distFromDX(member->getWaypointDX(), member->getWaypointDY())));
    y += h;

    // Line 5, 6: ETA and fuel usage
    ctx.setColor(UIColor::tc_Static);
    outTextF(ctx, x0, y, lwidth, tab(3));
    outTextF(ctx, x0, y+h, lwidth, tab(4));

    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    std::auto_ptr<GShipTurnPredictor> main, towee;
    if (member->getMission() == GMission::msn_Tow && univ.ty_played_ships.isValidIndex(member->getTowId())) {
        towee.reset(new GShipTurnPredictor(univ, member->getTowId()));
        main.reset(new GShipTurnPredictor(univ, member->getId(), *towee));
    } else {
        main.reset(new GShipTurnPredictor(univ, member->getId()));
    }
    main->computeMovement();

    if (member->getWaypointDX() == 0 && member->getWaypointDY() == 0) {
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, x1, y, w, _("at waypoint"));
    } else if (member->getWarp() == 0) {
        ctx.setColor(UIColor::tc_Red);
        outTextF(ctx, x1, y, w, _("not moving"));
    } else if (main->isAtTurnLimit()) {
        ctx.setColor(UIColor::tc_Yellow);
        outTextF(ctx, x1, y, w, _("too long"));
    } else {
        ctx.setColor(UIColor::tc_Green);
        outTextF(ctx, x1, y, w, format(_("%d turn%!1{s%}"), main->getNumTurns()));
    }

    if (main->isAtTurnLimit()) {
        ctx.setColor(UIColor::tc_Yellow);
    } else if (main->getMovementFuelUsed() > member->getCargo(el_Neutronium)) {
        ctx.setColor(UIColor::tc_Red);
    } else {
        ctx.setColor(UIColor::tc_Green);
    }
    // FIXME: cloak fuel, turn fuel
    outTextF(ctx, x1, y+h, w, format(_("%d kt"), numToString(main->getMovementFuelUsed())));
}
