/**
  *  \file client/widgets/objcontrol.cc
  *  \brief WObjectSelectionControl widget
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/widgets/objcontrol.h"
#include "game/obj.h"
#include "game/objtypes.h"
#include "game/univ.h"

/** Find object at location, with wrap
    \param type Object type to search in
    \param i Starting index
    \param forward true: browse forward, false: browse backward
    \param marked true: only marked items, false: all items
    \param pt Accept only units at this location
    \return new index; 0 if none found */
static int
findSameLocation(GObjectType& type, int i, bool forward, bool marked, GPoint pt)
{
    bool wrapped = false;
    while (1) {
        // Go one step
        if (forward) {
            i = type.findNextIndexNoWrap(i, marked);
        } else {
            i = type.findPreviousIndexWrap(i, marked);
        }

        // Terminate?
        if (i == 0) {
            // Reached end; note that and exit after reaching it a second time.
            if (wrapped) {
                break;
            }
            wrapped = true;
        } else {
            // Found some object; exit if it matches.
            if (type.isValidIndex(i)) {
                GMapObject* mo = dynamic_cast<GMapObject*>(&type.getObjectByIndex(i));
                if (mo != 0 && mo->getPos() == pt) {
                    break;
                }
            }
        }
    }

    return i;
}

WObjectSelectionControl::WObjectSelectionControl(GObjectSelection& selection)
    : UIFixedWidget(),
      selection(selection)
{ }

bool
WObjectSelectionControl::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_PAGEUP:
         case SDLK_UP:
         case vk_WheelUp:
         case '-':
            if (int i = selection.getObjectType()->findPreviousIndexWrap(selection.getCurrentIndex(), false)) {
                selection.setCurrentIndex(i);
            }
            return true;
         case SDLK_PAGEDOWN:
         case SDLK_DOWN:
         case vk_WheelDown:
         case '+':
            if (int i = selection.getObjectType()->findNextIndexWrap(selection.getCurrentIndex(), false)) {
                selection.setCurrentIndex(i);
            }
            return true;
         case ss_Ctrl + SDLK_PAGEUP:
         case ss_Ctrl + SDLK_UP:
            if (int i = selection.getObjectType()->findPreviousIndexWrap(selection.getCurrentIndex(), true)) {
                selection.setCurrentIndex(i);
            }
            return true;
         case ss_Ctrl + SDLK_PAGEDOWN:
         case ss_Ctrl + SDLK_DOWN:
            if (int i = selection.getObjectType()->findNextIndexWrap(selection.getCurrentIndex(), true)) {
                selection.setCurrentIndex(i);
            }
            return true;
         case SDLK_HOME:
            if (int i = selection.getObjectType()->findNextIndexNoWrap(0, false)) {
                selection.setCurrentIndex(i);
            }
            return true;
         case SDLK_END:
            if (int i = selection.getObjectType()->findPreviousIndexNoWrap(0, false)) {
                selection.setCurrentIndex(i);
            }
            return true;
         case ss_Ctrl + SDLK_HOME:
            if (int i = selection.getObjectType()->findNextIndexNoWrap(0, true)) {
                selection.setCurrentIndex(i);
            }
            return true;
         case ss_Ctrl + SDLK_END:
            if (int i = selection.getObjectType()->findPreviousIndexNoWrap(0, true)) {
                selection.setCurrentIndex(i);
            }
            return true;
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
         case ss_Ctrl + SDLK_TAB:
         case ss_Ctrl + ss_Shift + SDLK_TAB:
            if (GMapObject* mo = dynamic_cast<GMapObject*>(selection.getCurrentObject())) {
                bool forward = (event.key.code & ss_Shift) == 0;
                bool marked = (event.key.code & ss_Ctrl) != 0;
                if (int i = findSameLocation(*selection.getObjectType(), selection.getCurrentIndex(), forward, marked, mo->getPos())) {
                    selection.setCurrentIndex(i);
                }
            }
            return true;
         case '.':
            if (GObject* obj = selection.getCurrentObject()) {
                obj->setIsMarked(!obj->isMarked());
                selection.getCurrentUniverse()->doScreenUpdates();
            }
            return true;
        }
    }
    return UIFixedWidget::handleEvent(event, second_pass);
}

void
WObjectSelectionControl::drawContent(GfxCanvas& /*can*/)
{ }
