/**
  *  \file game/actions/planetbase.cc
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/actions/planetbase.h"
#include "game/planet.h"
#include "game/config.h"
#include "util/nls.h"
#include "game/actions/preconditions.h"

/*********************** GPlanetBuildStarbaseAction **********************/

/** \class GPlanetBuildStarbaseAction
    \brief Building and Scraping Starbases

    This class encapsulates the function to build a starbase, or to
    cancel such an order. */

/** Constructor.
    \param the_planet    where we want to build
    \param container     pays that bill
    \param want_build    true to build a base, false to scrap it (=cancel the build order)

    \pre We are playing the planet and the container */
GPlanetBuildStarbaseAction::GPlanetBuildStarbaseAction(GPlanet& the_planet,
                                                       GCargoContainer& container,
                                                       const bool want_build)
    : GCargoCostTransaction(container),
      the_planet(the_planet),
      want_build(want_build),
      conn(the_planet.sig_changed.add(this, &GPlanetBuildStarbaseAction::update))
{
    mustBePlayed(the_planet);
    if (want_build == the_planet.isBuildingBase()) {
        // null operation
        if (want_build)
            throw GError(GError::eDone, _("Starbase building already in progress"));
        else
            throw GError(GError::eDone, _("No starbase being built"));
    } else if (the_planet.hasBase()) {
        // base present
        throw GError(GError::ePerm, _("Starbase already present"));
    }

    update();
}

/** Update. Recomputes the cost. */
void
GPlanetBuildStarbaseAction::update()
{
    if (!isFinished()) {
        int factor = want_build - the_planet.isBuildingBase();
        setCost(config.StarbaseCost(the_planet.getOwner()) * factor);
    }
}

/** Destructor. */
GPlanetBuildStarbaseAction::~GPlanetBuildStarbaseAction()
{
    /* nothing to do, the GCargoContainer will do everything required */
}

void
GPlanetBuildStarbaseAction::commit()
{
    ASSERT(isValidCost());
    the_planet.setBuildBaseFlag(want_build);
    GCargoCostTransaction::commit();
}
