/**
  *  \file game/drawing.h
  *  \brief User Drawings
  *
  *  Classes to represent and manage user-created starchart drawings.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_DRAWING_H
#define PCC_V2_GAME_DRAWING_H

#include <map>
#include "game/coord.h"
#include "util/atom.h"
#include "cpluslib/string.h"
#include "cpluslib/types.h"
#include "cpluslib/iterator.h"
#include "cpluslib/sigslot.h"
#include "cpluslib/ptrmultilist.h"

class Stream;

/** Kind of drawing. This matches the "type of element" field in
    chartX.cc. */
enum GDrawingType {
    LineDrawing      = 0,       ///< Line from A to B.
    RectangleDrawing = 1,       ///< Rectangle between A and B.
    CircleDrawing    = 2,       ///< Circle with center and radius.
    MarkerDrawing    = 3        ///< Marker (tiny symbol) with optional comment; lockable.
};

/** User Drawing. This class represents a single user drawing
    object.

    \todo should we derive this from GMapObject? */
class GDrawing {
    GPoint pos;                 ///< Center position or left/top.
    GDrawingType type;          ///< Kind of drawing.
    uint8 color;                ///< Color.
    int x2, y2;                 ///< Radius, kind, or bottom/right.
    atom_t tag;                 ///< User-defined marking.
    int expire;                 ///< Turn of expiry.
    string_t comment;           ///< User comment.

    friend class GDrawingContainer; // This simplifies GDrawingContainer::save a little

 public:
    GDrawing(GPoint pos, GDrawingType type);

    void setPos(GPoint pos);
    void setPos2(GPoint pos);
    void setCircleRadius(int r);
    void setMarkerKind(int k);
    void setTag(atom_t tag);
    void setColor(uint8 color);
    void setComment(string_t comment);
    void setExpire(int expire);

    GDrawingType getType() const
        { return type; }
    GPoint getPos() const
        { return pos; }
    GPoint getPos2() const
        { return GPoint(x2, y2); }
    int getCircleRadius() const
        { return x2; }
    int getMarkerKind() const
        { return x2; }
    atom_t getTag() const
        { return tag; }
    uint8 getColor() const
        { return color; }
    string_t getComment() const
        { return comment; }
    int getExpire() const
        { return expire; }

    bool isVisible() const
        { return color != 0; }

    double getDistanceTo(GPoint pt) const;
    double getDistanceToWrap(GPoint pt) const;

    static int getUserColorFromColor(uint8 color);
    static uint8 getColorFromUserColor(int user_color);

    enum { NUM_USER_COLORS = 30 };
};

/** Drawing Tag Map for Save/Reload. Markers have an associated
    numeric tag, which can be an atom. The scripting interface
    guarantees that these atoms survive save/reload. This class
    supports save/reload by providing a mapping between the internal
    atom value and the externally-stored one. This class can also help
    in generating a list of used tags for filtering or drawing.

    Note that PCC 1.x always saves the internal atom numbers and
    remaps only on reload. We also remap on save, to compact our
    potentially larger range of atoms. */
class GDrawingAtomMap {
    typedef std::map<atom_t,int16> map_t;

    map_t atoms;
    int16 counter;

    enum { EXTERNAL_ATOM_MIN = 20000 };
 public:
    typedef member_iterator<map_t::const_iterator,
                            const atom_t,
                            &map_t::const_iterator::value_type::first> iterator;

    GDrawingAtomMap();
    ~GDrawingAtomMap();

    void clear();
    bool isEmpty() const;

    void add(atom_t a);

    void save(Stream& out);
    void load(Stream& in);

    iterator begin() const;
    iterator end() const;

    atom_t get(int16 value) const;
    int16 getExternalValue(atom_t atom) const;
};

/** Drawing container. Contains a list of drawings, and provides a set
    of operations on all of them, as well as iterators. */
class GDrawingContainer {
    typedef ptr_multi_list<GDrawing> list_t;
    list_t drawings;
 public:
    /// Flags for computeAtomMap().
    enum {
        IncludeNumeric   = 1, ///< Include numeric tags (default: only atoms). For filtering.
        IncludeInvisible = 2  ///< Include invisible markers (default: only visible). For saving.
    };
    typedef list_t::iterator iterator, const_iterator;

    GDrawingContainer();
    ~GDrawingContainer();

    iterator add(const GDrawing& drawing);

    void computeAtomMap(GDrawingAtomMap& map, int flags) const;
    void save(Stream& out, const GDrawingAtomMap& map) const;
    void load(Stream& in, const GDrawingAtomMap& map, int expire);

    iterator findNearestVisibleDrawing(GPoint pt);

    const_iterator begin() const
        { return drawings.begin(); }
    const_iterator end() const
        { return drawings.end(); }

    void erase(iterator it);
    void eraseExpiredDrawings(int turn);

    Signal0           sig_changed;
    Signal1<iterator> sig_deleted;
};

#endif
