/**
  *  \file game/sim-run.cc
  *  \brief Simulator Engine
  *
  *  This contains the routines to run a simulation using classic
  *  VCRs as well as FLAK.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <vector>
#include <algorithm>
#include "game/sim-run.h"
#include "game/fcode.h"
#include "game/config.h"
#include "util/random.h"
#include "game/formula.h"
#include "cpluslib/misc.h"
#include "defs.h"
#include "util/recode.h"
#include "game/classicvcr.h"

#ifdef CONF_FLAK_SUPPORT
# include "flak/flak.h"
# include "flak/flakconf.h"
# include "flak/flak-db.h"
#endif

static int
plimit(int max, int scale, int d)
{
    return max - (max * d) / scale;
}


static int
getSeed(const GSimOptions& opts, const GSimBattleResult& result)
{
    if (opts.seed_control) {
        if (opts.vcr_mode == opts.VcrNuHost) {
            return (result.this_battle_index % 118) + 1;
        } else {
            return (result.this_battle_index % 110) + 1;
        }
    } else {
        if (opts.vcr_mode == opts.VcrHost)
            return myRandom(110) + 1;
        else if (opts.vcr_mode == opts.VcrNuHost)
            return myRandom(118) + 1;
        else
            return myRandom();
    }
}

static int
getDamageTech(int tech, int damage)
{
    int max = (100 - damage) / 10;
    if (tech > max)
        tech = max;
    if (tech <= 0)
        tech = 1;
    return tech;
}

/** Bonus fighter table for Master at Arms compensation.
    First index is 0 for carrier/carrier, 1 for planet/carrier.
    Second index is number of effective bays of right carrier, plus 1.
    Third index is number of effective bays of left carrier, plus 1.
    (Effective bays has a range of -1 to 13).
    Value is average number of bonus fighters times 10, i.e. 64 means
    we're getting 6 bonus fighters, and with 40% probability, we get
    another bonus fighter. */
static const uint8 master_bonus_fighters_X10[2][15][15] = {
    //-1   0   1   2   3   4   5   6   7   8   9  10  11  12  13
   {{  0,  0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },  // -1
    {  1,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2 },  // 0
    {  3,  3,  3,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7,  7 },  // 1
    {  4,  7,  9, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19, 19 },  // 2
    {  4,  7,  9, 21, 19, 24, 26, 26, 27, 27, 27, 27, 27, 27, 27 },  // 3
    {  4,  7,  9, 21, 26, 35, 38, 40, 40, 40, 40, 40, 40, 40, 40 },  // 4
    {  4,  7,  9, 25, 33, 40, 44, 48, 52, 53, 53, 53, 53, 53, 53 },  // 5
    {  4,  7,  9, 25, 34, 45, 50, 52, 56, 58, 61, 64, 64, 64, 64 },  // 6
    {  4,  7,  9, 25, 37, 50, 54, 58, 62, 64, 67, 72, 75, 75, 75 },  // 7
    {  4,  7,  9, 25, 37, 53, 60, 66, 69, 73, 75, 78, 82, 82, 82 },  // 8
    {  4,  7,  9, 25, 37, 53, 63, 69, 71, 76, 79, 82, 86, 89, 90 },  // 9
    {  4,  7,  9, 25, 37, 53, 63, 71, 74, 78, 82, 86, 94, 95, 96 },  // 10
    {  4,  7,  9, 25, 37, 53, 63, 71, 80, 82, 84, 89, 98, 99,100 },  // 11
    {  4,  7,  9, 25, 38, 53, 63, 71, 80, 85, 89, 93, 99,101,104 },  // 12
    {  4,  7,  9, 25, 38, 53, 63, 71, 80, 85, 89, 94, 99,102,106 }}, // 13

   {{  0,  0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
    {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
    {  2,  2,  2,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4 },
    {  2,  4,  5, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10, 10 },
    {  2,  4,  5, 11, 10, 12, 13, 13, 14, 14, 14, 14, 14, 14, 14 },
    {  2,  4,  5, 11, 13, 18, 19, 20, 20, 20, 20, 20, 20, 20, 20 },
    {  2,  4,  5, 13, 17, 20, 22, 24, 26, 27, 27, 27, 27, 27, 27 },
    {  2,  4,  5, 13, 17, 23, 25, 26, 28, 29, 31, 32, 32, 32, 32 },
    {  2,  4,  5, 13, 19, 25, 27, 29, 31, 32, 34, 36, 38, 38, 38 },
    {  2,  4,  5, 13, 19, 27, 30, 33, 35, 37, 38, 39, 41, 41, 41 },
    {  2,  4,  5, 13, 19, 27, 32, 35, 36, 38, 40, 41, 43, 45, 45 },
    {  2,  4,  5, 13, 19, 27, 32, 36, 37, 39, 41, 43, 47, 48, 48 },
    {  2,  4,  5, 13, 19, 27, 32, 36, 40, 41, 42, 45, 49, 50, 50 },
    {  2,  4,  5, 13, 19, 27, 32, 36, 40, 43, 45, 47, 50, 51, 52 },
    {  2,  4,  5, 13, 19, 27, 32, 36, 40, 43, 45, 47, 50, 51, 53 }}
};

/** Bonus bay table for Master at Arms compensation.
    First index is 0 for carrier/carrier, 1 for planet/carrier.
    Second index is number of effective bays of right carrier, plus 1.
    Third index is number of effective bays of left carrier, plus 1.
    (Effective bays has a range of -1 to 13).
    Value is average number of bonus bays times 100, i.e. 108 means
    we receive 1 bonus bay, plus another one with 8% probability. */
static const uint8 master_bonus_bays_X100[2][15][15] = {
    //-1   0   1   2   3   4   5   6   7   8   9  10  11  12  13
   {{  0,  0,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },  // -1
    {  1,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2,  2 },  // 0
    {  2,  2,  2,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5,  5 },  // 1
    {  3,  5,  7, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14, 14 },  // 2
    {  3,  5,  7, 16, 14, 18, 20, 20, 20, 20, 20, 20, 20, 20, 20 },  // 3
    {  3,  5,  7, 16, 20, 27, 29, 30, 30, 30, 30, 30, 30, 30, 30 },  // 4
    {  3,  6,  7, 20, 26, 32, 35, 38, 41, 42, 42, 42, 42, 42, 42 },  // 5
    {  4,  7,  9, 24, 33, 43, 48, 50, 54, 56, 59, 62, 62, 62, 62 },  // 6
    {  5,  8, 10, 28, 42, 57, 61, 66, 71, 73, 76, 82, 85, 85, 85 },  // 7
    {  5,  9, 12, 33, 49, 70, 79, 87, 91, 96, 99,103,108,108,108 },  // 8
    {  6, 10, 13, 37, 55, 79, 94,103,106,114,118,123,129,133,135 },  // 9
    {  7, 12, 15, 42, 62, 89,106,120,125,131,138,145,158,160,162 },  // 10
    {  8, 13, 17, 47, 69, 99,118,133,150,154,158,167,184,186,188 },  // 11
    {  8, 15, 19, 52, 79,110,131,147,166,176,185,193,205,209,216 },  // 12
    {  9, 16, 21, 57, 87,121,144,162,182,194,203,214,226,232,241 }}, // 13

   {{  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
    {  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0,  0 },
    {  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1,  1 },
    {  1,  1,  2,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4,  4 },
    {  1,  1,  2,  4,  4,  4,  5,  5,  5,  6,  6,  6,  6,  6,  6 },
    {  1,  1,  2,  4,  5,  7,  8,  8,  8,  8,  8,  8,  8,  8,  8 },
    {  1,  2,  2,  6,  8,  9, 10, 11, 12, 12, 12, 12, 12, 12, 12 },
    {  1,  2,  3,  7, 10, 13, 14, 15, 16, 17, 17, 18, 18, 18, 18 },
    {  1,  2,  3,  8, 12, 17, 18, 20, 21, 22, 23, 24, 25, 25, 25 },
    {  2,  3,  4, 10, 14, 21, 23, 26, 27, 29, 29, 30, 32, 32, 32 },
    {  2,  3,  4, 11, 16, 24, 28, 31, 32, 34, 35, 37, 38, 40, 40 },
    {  2,  4,  5, 13, 19, 27, 32, 36, 37, 39, 41, 43, 47, 48, 48 },
    {  2,  4,  5, 14, 21, 30, 35, 40, 45, 46, 47, 50, 55, 55, 56 },
    {  2,  4,  6, 16, 24, 33, 39, 44, 50, 53, 55, 58, 61, 63, 64 },
    {  3,  5,  6, 17, 26, 36, 43, 49, 55, 58, 61, 64, 68, 70, 72 }}
};

/**************************** GSimBattleResult ***************************/

/** Initialize.
    \param this_battle_index Index of the battle to be simulated, 0-based. */
void
GSimBattleResult::init(const GSimOptions& opts, int this_battle_index)
{
    this->this_battle_index   = this_battle_index;
    this->this_battle_weight  = 1;
    this->total_battle_weight = 1;
    this->series_length       = opts.vcr_mode == opts.VcrNuHost ? 118 : 110;
    this->battles             = 0;
}

/** Add a series of a given length. Returns the position in the
    series. For example, "addSeries(2)" says that we double the length
    of a series because there are two cases to handle, and returns 0
    or 1 saying in which case we are. */
int
GSimBattleResult::addSeries(int length)
{
    int result = this_battle_index / series_length;
    series_length *= length;
    return result % length;
}

/** Adjust weight. Force this battle's weight to the given value. Use
    this to make the counters comparable to another GSimBattleResult
    with the specified weight.
    \param new_weight New weight */
void
GSimBattleResult::changeWeightTo(int32 new_weight)
{
    this_battle_weight = this_battle_weight * new_weight / total_battle_weight;
    total_battle_weight = new_weight;
}

/**************************** HOST Simulation ****************************/

static int
getFCodeValueTHost(const GSimShip& a)
{
    /* FIXME: this uses the host's fcode parser, not the one
       configured for battle simulation. We fix up biggest mistakes by
       dropping '-99' codes, but hasSpacePaddedFCodes() still remains
       as not 100% parsed option. */
    int val = GFCode::getNumericValue(a.getFCode());
    if (val <= 0)
        val = 1000;
    if (val == 1000) {
        if (a.getAggressiveness() == GSimShip::agg_NoFuel || a.getAggressiveness() == GSimShip::agg_Passive)
            val += 15;
    }
    return val;
}

bool
sortByBattleOrderTHost(const GSimShip* a, const GSimShip* b)
{
    int abo = getFCodeValueTHost(*a);
    int bbo = getFCodeValueTHost(*b);
    if (abo != bbo)
        return abo < bbo;
    return a->getId() < b->getId();
}

/** Predicate to sort by Id number, backwards. Used for Tim-Host
    intercept attack. */
static bool
sortByIdBackwards(const GSimShip* a, const GSimShip* b)
{
    return a->getId() > b->getId();
}

static int
getFCodeValuePHost(GSimObject& a)
{
    int val = GFCode::getNumericValue(a.getFCode());
    if (val >= 1000) {
        if (GSimPlanet* p = dynamic_cast<GSimPlanet*>(&a)) {
            if (config.AllowPlanetAttacks() && (a.getFCode() == "ATT" || a.getFCode() == "NUK"))
                val = 0;
            else if (p->getDefense() != 0)
                val = 1001;
            else
                val = 1003;
        } else if (GSimShip* s = dynamic_cast<GSimShip*>(&a)) {
            if (s->getBeamCount() != 0 || s->getTorpLauncherCount() != 0 || s->getBayCount() != 0)
                val = 1000;
            else
                val = 1002;
        }
    }
    return val;
}

bool
sortByBattleOrderPHost(GSimObject* a, GSimObject* b)
{
    int abo = getFCodeValuePHost(*a);
    int bbo = getFCodeValuePHost(*b);
    if (abo != bbo)
        return abo < bbo;
    if (a->getId() != b->getId())
        return a->getId() < b->getId();
    return dynamic_cast<GSimShip*>(a) != 0;
}

/** Check whether two ships attack.
    \param at   potential attacker
    \param op   potential opponent
    \param opts simulation options
    \return true iff at attacks op. Checks only one direction!

    ccSim.pas, IsAttacking */
static bool
isAttacking(const GSimShip& at, const GSimShip& op, const GSimOptions& opts)
{
    /* Conditions according to host.exe:
       no fight if players are the same                             (1)
       no fight if right ship offers alliance to left ship          (2)
       if CloakedShipsAttack=0:
         no fight if any ship cloaked                               (3)
       else
         no fight if both ships cloaked                             (4)
         no fight if PE of cloaked ship does not match other ship   (5)
       no fight if either ship already dead                         (6)
       no fight if either ship is fuelless                          (7)
       no fight if neither ship has beams nor bays                  (8)
       no fight if fcode match and fcode not lfm,NTP,mkt            (9)
       no fight if either ship has mission 0 and 10 crew            (10)

       otherwise, fight happens if there's a PE match or mission 4  (11)

       Deviations: (2) is handled symmetrically; should probably test
       it and if it's relevant, implement it. (8) is implemented as
       real freighter test. It remains to test whether (10) is
       relevant.

       Note that (5) implicitly follows from (11), so we don't check
       that explicitly. */

    /* Conditions according to PHost:
       no fight if either ship already dead                         (p1)  (=6)
       no fight if aggressor has no fuel                            (p2)  (=7)
       no fight if aggressor has neither PE nor mission 4           (p3)  (=11)
       no fight if players are the same                             (p4)  (=1)
       no fight if aggressor offers alliance to opponent            (p5)  (=2)
       no fight if opponent has no fuel                             (p6)  (=7)
       if CloakedShipsAttack=0:
         no fight if either ship is cloaked                         (p7)  (=3)
       else
         no fight if both ships cloaked                             (p8)  (=4)
         no fight if opponent is cloaked and has mismatching PE     (p9)
       no fight if FC match                                         (p11) (=9)

       These are almost the same rules as HOST. (5) follows from (11)
       = (p3), so it need not be implemented. (8) is not implemented
       in PHost up to 4.0j(!!), which is considered a bug. */

    /* deactivated units do not fight */
    if (((at.getFlags() | op.getFlags()) & GSimShip::fl_Deactivated) != 0)
        return false;
    /* same owner does not fight */
    if (at.getOwner() == op.getOwner())
        return false;
    /* zombies do not fight */
    if (at.getOwner() == 0 || op.getOwner() == 0)
        return false;
    /* friends do not fight */
    if (opts.honor_alliances && opts.alliance_settings.get(at.getOwner(), op.getOwner()))
        return false;
    /* passive or fuelless units do not attack */
    if (at.getAggressiveness() == at.agg_Passive || at.getAggressiveness() == at.agg_NoFuel)
        return false;
    /* we have a PE, so check whether opponent matches */
    if (at.getAggressiveness() != at.agg_Kill) {
        if (op.getOwner() != at.getAggressiveness() && !opts.enemy_settings.get(at.getOwner(), op.getOwner()))
            return false;
    }
    /* check for cloaking */
    if ((at.getFlags() & GSimShip::fl_Cloaked) != 0) {
        if (!config.AllowCloakedShipsAttack())
            return false;
    }
    /* check whether enemy is cloaked */
    /* FIXME: PHost has silly exception here: you *can* attack a cloaked
       ship if they have you as their enemy */
    if ((op.getFlags() & GSimShip::fl_Cloaked) != 0)
        return false;

    /* check for fuel */
    if (op.getAggressiveness() == op.agg_NoFuel)
        return false;

    /* Now check friendly codes */
    string_t afc = at.getFCode();
    if (afc == op.getFCode()) {
        if (afc != "mkt" && afc != "lfm" && afc != "NTP" && afc != "???" && !GFCode::isExtraFC(afc))
            return false;
    }

    return true;
}

/** Check whether a ship is immune from planet attacks.
    \param sh    [in] the ship
    \return true iff ship is immune
    \todo PHost 4.0j hull functions */
static bool
isImmune(const GSimShip& sh)
{
    /* Conditions in Host 3.22.40:
       - immune if owned by Rebel and !PlanetsAttackRebels [handled by hasShipFunction]
       - immune if owned by Klingons and !PlanetsAttackKlingons [handled by hasShipFunction]
       - immune if owned by Birds, has no fuel, but has beams
       - immune if it is a SSD
       - immune if ship is cloaked */
    if (raceId(sh.getOwner()) == 3 && sh.getAggressiveness() == sh.agg_NoFuel && sh.getBeamCount() != 0)
        return true;
    if (sh.hasFunction(sf_PlanetImmunity))
        return true;
    if (sh.getFlags() & sh.fl_Cloaked)
        return true;

    return false;
}

/** Check whether a ship/planet attack each other. Unlike the version
    which takes two GSimShip, this one tests both directions of
    aggression. The two-GSimShip version tests only one.
    \param left  [in] Ship
    \param right [in] Planet
    \param opts  [in] Simulation options */
static bool
isAttacking(const GSimShip& left, const GSimPlanet& right, const GSimOptions& opts)
{
    /* Host 3.22.40:
       - if planet.fc="NUK" && AllowPlanetAttacks && planet.defense=0 && ship.owner=5
         then planet.defense <- 1

       no fight if ship is cloaked                           (1)
       fight if ship is aggressive:
         no alliance offer                                   (2)
         and mission kill or matching PE                     (3)
         and has fuel                                        (4)
         and not fcode match                                 (5)

       fight if planet is aggressive:
         no alliance offer                                   (6)
         and ship not immune                                 (7)
         and "ATT", AllowPlanetAttacks, ship has fuel        (8)
             or "NUK", AllowPlanetAttacks, planet has defense */

    /* PHost differentiates between aggressor and opponent:

       if planet is aggressor:
         no fight if neither ATT nor NUK, or                 (p1) = (8)
           !AllowPlanetAttacks
         no fight if ship is immune                          (p2) = (7)
         no fight if ship is fuel-less and (fcode not NUK    (p3) = (7)
            or fuelless capital bird ship)

       if planet is opponent:
         no fight if out of fuel                             (p4) = (4)
         no fight if neither kill nor PE                     (p5) = (3)
         no fight if PE mismatch                             (p6) = (3)

       no fight if same owner                                (p7)
       no fight if aggressor offers alliance                 (p8) = (2)+(6)
       no fight if ship is cloaked                           (p9) = (7)+(1)
       no fight if fcode match                               (p10) = (5) */

    /* deactivated units don't fight */
    if (((left.getFlags() | right.getFlags()) & GSimObject::fl_Deactivated) != 0)
        return false;
    /* same owner does not fight */
    if (left.getOwner() == right.getOwner())
        return false;
    /* zombies do not fight */
    if (left.getOwner() == 0 || right.getOwner() == 0)
        return false;
    /* cloaked ships do not fight */
    if ((left.getFlags() & GSimObject::fl_Cloaked) != 0)
        return false;
    /* same FCode does not fight */
    string_t fc = left.getFCode();
    if (fc == right.getFCode()) {
        if (fc != "ATT" && fc != "NUK" && fc != "???" && !GFCode::isExtraFC(fc))
            return false;
    }

    /* does the ship want to attack the planet? */
    bool ship_wants_attack;
    if (opts.honor_alliances && opts.alliance_settings.get(left.getOwner(), right.getOwner()))
        ship_wants_attack = false;
    else if (left.getAggressiveness() == left.agg_Kill)
        ship_wants_attack = true;
    else if (left.getAggressiveness() == left.agg_NoFuel || left.getAggressiveness() == left.agg_Passive)
        ship_wants_attack = false;
    else if (left.getAggressiveness() == right.getOwner() || opts.enemy_settings.get(left.getOwner(), right.getOwner()))
        ship_wants_attack = true;
    else
        ship_wants_attack = false;

    /* does the planet want to attack the ship? */
    /* FIXME: handle the "1 dp" condition, and the anti-NUK-trap rule */
    bool planet_wants_attack;
    if (opts.honor_alliances && opts.alliance_settings.get(right.getOwner(), left.getOwner()))
        planet_wants_attack = false;
    else if (right.getFCode() == "ATT")
        planet_wants_attack = (left.getAggressiveness() != left.agg_NoFuel);
    else if (right.getFCode() == "NUK")
        planet_wants_attack = true;
    else
        planet_wants_attack = false;

    return ship_wants_attack
        || (planet_wants_attack && !isImmune(left));
}

/** Check whether any two objects attack each other. Unlike that
    isAttacking() functions, this one can take any object combination
    in any order. */
static bool
isAttackingAny(const GSimObject& a, const GSimObject& b, const GSimOptions& opts)
{
    const GSimShip* as   = dynamic_cast<const GSimShip*>(&a);
    const GSimShip* bs   = dynamic_cast<const GSimShip*>(&b);
    const GSimPlanet* ap = dynamic_cast<const GSimPlanet*>(&a);
    const GSimPlanet* bp = dynamic_cast<const GSimPlanet*>(&b);

    if (as && bs)
        return isAttacking(*as, *bs, opts) || isAttacking(*bs, *as, opts);
    if (as && bp)
        return isAttacking(*as, *bp, opts);
    if (ap && bs)
        return isAttacking(*bs, *ap, opts);
    return false;
}

/** Check whether ship is armed. */
static bool
isArmed(const GSimShip& sh)
{
    return sh.getBeamCount() != 0
        || sh.getTorpLauncherCount() != 0
        || sh.getBayCount() != 0;
}

/** Pack ship into VCR record.
    \param obj  [out] VCR side
    \param sh   [in] Ship to pack */
static void
packShip(GVcrObject& obj, const GSimShip& sh, const GSimOptions& opts)
{
    obj.setIsPlanet(false);
    if (sh.getBayCount() != 0) {
        obj.setFighterCount(sh.getAmmo());
        obj.setTorpCount(0);
        obj.setTorpLauncherCount(0);
        obj.setBayCount(sh.getBayCount());
        obj.setTorpType(0);
    } else if (sh.getTorpLauncherCount() != 0) {
        obj.setFighterCount(0);
        obj.setTorpCount(sh.getAmmo());
        obj.setTorpLauncherCount(sh.getTorpLauncherCount());
        obj.setBayCount(0);
        obj.setTorpType(sh.getTorpType());
    } else {
        obj.setFighterCount(0);
        obj.setTorpCount(0);
        obj.setTorpLauncherCount(0);
        obj.setBayCount(0);
        obj.setTorpType(0);
    }
    obj.setName(sh.getName());
    obj.setDamage(sh.getDamage());
    obj.setCrew(sh.getCrew());
    obj.setId(sh.getId());
    obj.setOwner(sh.getOwner());
    obj.setRace(raceId(sh.getOwner()));
    obj.setBeamKillRate(raceId(sh.getOwner()) == 5 ? 3 : 1);
    obj.setBeamType(sh.getBeamType());
    obj.setBeamCount(sh.getBeamCount());
    obj.setExperienceLevel(opts.isExperienceEnabled() ? sh.getExperienceLevel() : 0);
    obj.setShield(sh.getShield());

    if (sh.getHull() != 0) {
        obj.setMass(getHull(sh.getHull()).getMass());
        obj.setPicture(getHull(sh.getHull()).getInternalPictureNumber());
        obj.setHull(sh.getHull());
    } else {
        obj.setMass(sh.getMass());
        obj.setPicture(200); // FIXME: hardcoded; this is the same value as used in PCC
        obj.setHull(0);
    }
}

/** Check whether friendly code states an ammunition limit.
    \param fc Friendly code
    \return Maximum number of torpedoes/fighters to use, -1 if no limit */
static int
getFCodeAmmoLimit(const string_t fc)
{
    // FIXME: move into FCode module?
    // FIXME: configuration
    if (fc.size() == 3 && fc[0] == 'N' && fc[1] == 'T') {
        if (fc[2] == 'P')
            return 0;                               /* NTP */
        else if (fc[2] == '0')
            return 100;                             /* NT0 */
        else if (fc[2] >= '1' && fc[2] <= '9')
            return 10*(fc[2] - '0');                /* NT1..NT9 */
        else
            return -1;
    } else {
        return -1;
    }
}

/** Unpack ship from VCR to simulation data.

    Note: this routine must not be called more than once on a given
    vcr/sh pair. Otherwise, torps might get lost when NTx is used.

    This routine also performs post-simulation modificators.

    \param obj  [in] VCR unit
    \param sh   [in/out] Target
    \param opts [in] Simulator options */
static void
unpackShip(const GVcrObject& obj, GSimShip& sh, const GSimOptions& opts)
{
    // Copy values
    sh.setShield(obj.getShield());
    sh.setDamage(obj.getDamage());
    sh.setCrew(obj.getCrew());

    // PHost, NuHost do fed bonus here
    if (raceId(sh.getOwner()) == 1 && opts.scotty_bonus) {
        if (opts.vcr_mode != GSimOptions::VcrHost) {
            sh.setShield(std::min(sh.getShield() + 25, 100));
        }
    }

    int limit = getFCodeAmmoLimit(sh.getFCode());
    int unused_ammo = 0;
    if (limit >= 0) {
        if (limit < sh.getAmmo())
            unused_ammo = sh.getAmmo() - limit;
    }
    if (sh.getBayCount())
        sh.setAmmo(obj.getFighterCount() + unused_ammo);
    else if (sh.getTorpLauncherCount())
        sh.setAmmo(obj.getTorpCount() + unused_ammo);
}

/** Apply ship modificators to VCR.
    - Engine-Shield Bonus
    - Bonus Bays
    - Fed Shield / Mass Bonus
    - no shields for freighters
    - damage limits
    - NTP
    \param obj  [in/out] VCR unit containing the ship.
    \param againstPlanet [in] true if opponent is a planet
    \param sh   [in] Original simulation ship data
    \param opts [in] Simulation options
    \pre all data belonging to the ship must be initialized. */
static void
applyShipModificators(GVcrObject& obj, bool againstPlanet, const GSimShip& sh, const GSimOptions& opts, const GPlayerArray<int>& level_base)
{
    /* engine-shield bonus */
    int bonus = 0;
    bool hosty = opts.vcr_mode == GSimOptions::VcrHost || opts.vcr_mode == GSimOptions::VcrNuHost;
    if (!againstPlanet || (!hosty && config.AllowESBonusAgainstPlanets()))
        bonus += opts.engine_shield_bonus;
    if (!hosty)
        bonus += getExperienceBonus(config.EModEngineShieldBonusRate, obj.getExperienceLevel());
    if (bonus != 0)
        // FIXME: HOST uses mass = ERND(mass + bonus*rate/100)
        obj.addMass(int32_t(bonus) * getEngine(sh.getEngineType()).getCost().get(el_Money) / 100);

    /* bonus bays and scotty bonus */
    bonus = 0;
    if (raceId(sh.getOwner()) == 1 && opts.scotty_bonus) {
        /* only THost bonus bays; PHost bonus bays handled below */
        if (hosty)
            bonus += 3;
        obj.addMass(50);
        if (opts.vcr_mode == GSimOptions::VcrHost) {
            // Only HOST, NuHost behaves like PHost here
            obj.setShield(obj.getShield() + 25);
        }
        obj.setShield(std::max(0, std::min(int(obj.getShield()), int(100 - obj.getDamage()))));
    }
    if (!hosty) {
        bonus += config.ExtraFighterBays(sh.getOwner());
        bonus += getExperienceBonus(config.EModExtraFighterBays, obj.getExperienceLevel());
    }
    if (obj.getBayCount() != 0) {
        obj.addBays(bonus);
        // Note that we need "bonus" later on for damage limits
    }

    /* Freighters have no shields */
    if (!isArmed(sh))
        obj.setShield(0);

    /* NTP & Co. */
    int limit = getFCodeAmmoLimit(sh.getFCode());
    if (limit >= 0) {
        if (obj.getFighterCount() > limit)
            obj.setFighterCount(limit);
        if (obj.getTorpCount() > limit)
            obj.setTorpCount(limit);
    }

    /* Damage limitations */
    if ((raceId(sh.getOwner()) != 1 || !opts.scotty_bonus)
        && !sh.hasFunction(sf_FullWeaponry))
    {
        if (hosty) {
            int limit = 10 - obj.getDamage() / 10;
            if (raceId(sh.getOwner()) == 2)
                limit += 5;
            if (limit < 0)
                limit = 0;
            obj.setTorpLauncherCount(std::min(obj.getTorpLauncherCount(), limit));
            obj.setBayCount         (std::min(obj.getBayCount(), limit));
            obj.setBeamCount        (std::min(obj.getBeamCount(), limit));
        } else {
            int limit = 100;
            if (raceId(sh.getOwner()) == 2)
                limit = 150;
            if (sh.getHull() == 0) {
                obj.setTorpLauncherCount(std::min(obj.getTorpLauncherCount(), plimit(sh.getTorpLauncherCount(), limit, sh.getDamage())));
                obj.setBayCount         (std::min(obj.getBayCount(),          plimit(sh.getBayCount() + bonus,  limit, sh.getDamage())));
                obj.setBeamCount        (std::min(obj.getBeamCount(),         plimit(sh.getBeamCount(),         limit, sh.getDamage())));
            } else {
                GHull& h = getHull(sh.getHull());
                obj.setTorpLauncherCount(std::min(obj.getTorpLauncherCount(), plimit(h.getMaxLaunchers(),    limit, sh.getDamage())));
                obj.setBayCount         (std::min(obj.getBayCount(),          plimit(h.getNumBays() + bonus, limit, sh.getDamage())));
                obj.setBeamCount        (std::min(obj.getBeamCount(),         plimit(h.getMaxBeams(),        limit, sh.getDamage())));
            }
        }
    }

    /* simplifications */
    if (obj.getTorpLauncherCount() == 0) {
        obj.setTorpCount(0);
        obj.setTorpType(0);
    }
    if (obj.getBayCount() == 0) {
        obj.setFighterCount(0);
    }
    if (obj.getBeamCount() == 0) {
        obj.setBeamType(0);
    }

    /* Finally, add level bonus */
    if (level_base.isValidIndex(sh.getOwner()) && obj.getExperienceLevel() < level_base[sh.getOwner()])
        obj.setExperienceLevel(obj.getExperienceLevel() + 1);

    /* Special abilities */
    obj.setBeamKillRate(sh.hasFunction(sf_TripleBeamKill) ? 3 : 1);
    obj.setBeamChargeRate(sh.hasFunction(sf_DoubleBeamCharge) ? 2 : 1);
    obj.setTorpChargeRate(sh.hasFunction(sf_DoubleTorpCharge) ? 2 : 1);
    obj.setCrewDefenseRate(sh.hasFunction(sf_Squadron) ? 100 : 0);
}

/** Apply modificators that apply to an opponent.
    \param obj  [in/out] VCR object to modify
    \param opp  [in] Its opponent
    \param opts [in] Options */
static void
applyOpponentModificators(GVcrObject& obj, const GSimShip& opp, const GSimOptions& /*opts*/)
{
    /* "Elusive" ability (Nu). Documented as "Ship has 10% rate of being hit by torpedoes",
       but the combat code only has a hit rate on the opponent. So I assume it is implemented
       this way: */
    if (opp.hasFunction(sf_Elusive)) {
        obj.setTorpMissRate(90);
    }
}

/** Apply planet modificators to VCR.
    - Commander level bonus
    - Special abilities
    \param obj  [in/out] VCR unit containing the planet
    \param pl   [in] Original simulation ship data
    \param level_base [in] Levels, for commanders
    \pre all data belonging to the planet must be initialized. */
static void
applyPlanetModificators(GVcrObject& obj, const GSimPlanet& pl, const GPlayerArray<int>& level_base)
{
    /* add level bonus */
    if (level_base.isValidIndex(pl.getOwner()) && obj.getExperienceLevel() < level_base[obj.getOwner()])
        obj.setExperienceLevel(obj.getExperienceLevel() + 1);

    /* Special abilities */
    obj.setBeamKillRate(pl.hasFunction(sf_TripleBeamKill) ? 3 : 1);
    obj.setBeamChargeRate(pl.hasFunction(sf_DoubleBeamCharge) ? 2 : 1);
    obj.setTorpChargeRate(pl.hasFunction(sf_DoubleTorpCharge) ? 2 : 1);
}

/** Apply Master at Arms bonus. Master at Arms is a proposal from
    Sirius (Jan Klingele) to balance combat in Tim-Host. It has, so
    far, not been implemented anywhere, but CCBSim has it (it used to
    be a proof-of-concept implementation of the proposal), so we offer
    it, too. The balancing approach attempts to fix the fighter
    intercept imbalance due to Tim's biased random number generator.
    \param vcr     [in/out] The prepared battle
    \param result  [in/out] Result meta-information (weight)
    \param opts    [in] Configuration */
static void
applyMasterBonus(GClassicVcr& vcr, GSimBattleResult& result, const GSimOptions& opts)
{
    /* only for fighter/fighter battles */
    if (vcr(0).getBayCount() == 0 || vcr(1).getBayCount() == 0)
        return;

    /* Compute maximum number of bonus fighters. This part of the
       algorithm is *not* in my paper / Winword copy of Master at
       Arms, and unfortunately, I cannot find my electronic
       communication with the Master at Arms people as well. However,
       this part is in PCC 1.x's ccsim.pas, so it probably wound up
       there after a suggestion by Sirius. The idea is to avoid giving
       a bonus if ships have so few fighters that fighter intercept is
       not an issue. */
    int eleft  = vcr(0).getFighterCount() - 2 * vcr(1).getBeamCount();
    int eright = vcr(1).getFighterCount() - 2 * vcr(0).getBeamCount();
    if (vcr(0).getShield() >= 100)
        eright -= vcr(0).getBeamCount();
    if (vcr(1).getShield() >= 100)
        eleft -= vcr(1).getBeamCount();

    /* we now have eleft, eright = effective fighters aboard
       (the fighters which do harm to the enemy, i.e. are not
       destroyed immediately after launch */
    int max_ef = std::max(0, std::min(eleft, eright));

    /* maximum bonus fighters 14% of max_ef. This computes
       10 times the maximum bonus fighters, and rounds it. */
    int max_bonus = (max_ef * 14 + 5) / 10;

    /* The following is straight from Master at Arms:
       Compute effective bay count. We add one to offset C indexing. */
    eleft  = vcr(0).getBayCount() - (vcr(1).getBeamCount() + 2) / 5 + 1;
    eright = vcr(1).getBayCount() - (vcr(0).getBeamCount() + 2) / 5 + 1;
    eleft  = std::max(0, std::min(eleft,  14));
    eright = std::max(0, std::min(eright, 14));

    /* Now apply the bonus */
    int bonus_bays_100    = master_bonus_bays_X100[vcr(1).isPlanet()][eright][eleft];
    int bonus_fighters_10 = master_bonus_fighters_X10[vcr(1).isPlanet()][eright][eleft];
    if (bonus_fighters_10 > max_bonus)
        bonus_fighters_10 = max_bonus;

    vcr(1).addBays(bonus_bays_100 / 100);
    vcr(1).addFighters(bonus_fighters_10 / 10);

    if (opts.seed_control) {
        if (result.addSeries(2)) {
            /* give bay bonus */
            vcr(1).addBays(1);
            result.this_battle_weight *= bonus_bays_100 % 100;
        } else {
            /* no bonus */
            result.this_battle_weight *= 100 - (bonus_bays_100 % 100);
        }
        result.total_battle_weight *= 100;

        /* FIXME: this can generate battles of weight 0 (PCC 1.x has
           the same problem). Solution (a): filter these out before
           running them. Solution (b): avoid generating them. This is
           not easy because input to this routine is not deterministic
           in case random left-right is used. */
        if (result.addSeries(2)) {
            /* give fighter bonus */
            vcr(1).addFighters(1);
            result.this_battle_weight *= bonus_fighters_10 % 10;
        } else {
            /* no bonus */
            result.this_battle_weight *= 10 - (bonus_fighters_10 % 10);
        }
        result.total_battle_weight *= 10;
    } else {
        if (myRandom(100) < bonus_bays_100 % 100)
            vcr(1).addBays(1);
        if (myRandom(10) < bonus_fighters_10 % 10)
            vcr(1).addFighters(1);
    }
}

/** Pack planet into VCR.
    \param obj [out] The VCR object
    \param pl  [in] The Planet */
static void
packPlanet(GVcrObject& obj, const GSimPlanet& pl, const GSimOptions& opts)
{
    const GSimOptions::VcrMode mode = opts.vcr_mode;
    const bool has_base = pl.getBaseBeamTech() > 0;

    const int planet_defense = pl.getDefense();
    const int base_defense   = has_base ? pl.getBaseDefense() : 0;
    const int base_fighters  = has_base ? pl.getBaseFighters() : 0;

    obj.setIsPlanet(true);
    if (mode != GSimOptions::VcrHost && mode != GSimOptions::VcrNuHost) {
        /* PHost */
        const long eff_p_defense = planet_defense * (100L - pl.getDamage()) / 100;
        const long eff_bp_defense = (planet_defense + base_defense) * (100L - pl.getDamage()) / 100;
        const int weapon_limit = config.AllowAlternativeCombat() ? 20 : 10;

        obj.setFighterCount(roundToInt(std::sqrt(double(eff_p_defense))) + base_fighters);
        obj.setTorpCount(0);
        obj.setTorpLauncherCount(0);
        obj.setName(pl.getName());
        obj.setDamage(pl.getDamage());
        obj.setCrew(0);
        obj.setId(pl.getId());
        obj.setOwner(pl.getOwner());
        obj.setRace(raceId(pl.getOwner()));
        obj.setBeamKillRate(raceId(pl.getOwner()) == 5 ? 3 : 1);
        obj.setPicture(1);
        obj.setHull(0);
        obj.setBeamType(roundToInt(std::sqrt(eff_p_defense / 2.0)));
        if (has_base && getDamageTech(pl.getBaseBeamTech(), pl.getBaseDamage()) > obj.getBeamType())
            obj.setBeamType(getDamageTech(pl.getBaseBeamTech(), pl.getBaseDamage()));
        if (obj.getBeamType() < 1)
            obj.setBeamType(1);
        if (obj.getBeamType() > NUM_BEAMS)
            obj.setBeamType(NUM_BEAMS);
        obj.setBeamCount(roundToInt(std::sqrt(eff_bp_defense / 3.0)));
        if (obj.getBeamCount() > weapon_limit)
            obj.setBeamCount(weapon_limit);
        obj.setExperienceLevel(opts.isExperienceEnabled() ? pl.getExperienceLevel() : 0);
        obj.setBayCount(obj.getFighterCount() - base_fighters);
        if (has_base)
            obj.addBays(5);
        obj.setTorpType(0);
        obj.setMass(100 + eff_p_defense + base_defense * (100L - pl.getDamage()) / 100);
        obj.setShield(pl.getShield());

        if (config.PlanetsHaveTubes()) {
            obj.setTorpType(roundToInt(std::sqrt(eff_p_defense / 2.0)));
            if (has_base && getDamageTech(pl.getBaseTorpTech(), pl.getBaseDamage()) > obj.getTorpType())
                obj.setTorpType(getDamageTech(pl.getBaseTorpTech(), pl.getBaseDamage()));
            if (obj.getTorpType() > NUM_TORPS)
                obj.setTorpType(NUM_TORPS);
            obj.setTorpLauncherCount(roundToInt(std::sqrt(eff_bp_defense / 4.0)));
            if (obj.getTorpLauncherCount() > 20) // FIXME: is this correct?
                obj.setTorpLauncherCount(20);

            /* planetary torps */
            int ppt = config.PlanetaryTorpsPerTube(obj.getOwner());
            ppt += getExperienceBonus(config.EModPlanetaryTorpsPerTube, obj.getExperienceLevel());

            obj.setTorpCount(ppt * obj.getTorpLauncherCount());

            /* add base storage torps */
            if (config.UseBaseTorpsInCombat(pl.getOwner())) {
                uint32 cost = 0;
                for (int i = 1; i <= NUM_TORPS; ++i)
                    cost += pl.getBaseTorps(i) * getTorp(i).getTorpCost();
                if (obj.getTorpType() > 0) {
                    int torp_cost = getTorp(obj.getTorpType()).getTorpCost();
                    if (torp_cost > 0)
                        obj.addTorpedoes(cost / torp_cost);
                }
            }
            if (obj.getTorpCount() > 255)
                obj.setTorpCount(255);
        }
    } else {
        /* Host */
        obj.setFighterCount(roundToInt(std::sqrt(double(planet_defense))) + base_fighters);
        obj.setTorpCount(0);
        obj.setTorpLauncherCount(0);

        obj.setName(pl.getName());

        obj.setDamage(0);
        obj.setCrew(obj.getFighterCount());                /* HOST does that, so we do it too,
                                                              in case we someday export the VCRs
                                                              to vcr.exe */
        obj.setId(pl.getId());
        obj.setOwner(pl.getOwner());
        obj.setRace(raceId(pl.getOwner()));
        obj.setBeamKillRate(raceId(pl.getOwner()) == 5 ? 3 : 1);
        obj.setPicture(1);
        obj.setHull(0);

        obj.setBeamType(roundToInt(std::sqrt(planet_defense / 2.0)));
        if (has_base && pl.getBaseBeamTech() > obj.getBeamType())
            obj.setBeamType(pl.getBaseBeamTech());
        if (obj.getBeamType() < 1)
            obj.setBeamType(1);
        if (obj.getBeamType() > NUM_BEAMS)
            obj.setBeamType(NUM_BEAMS);

        obj.setBeamCount(roundToInt(std::sqrt((planet_defense + base_defense) / 3.0)));
        if (obj.getBeamCount() > 10)
            obj.setBeamCount(10);

        obj.setExperienceLevel(opts.isExperienceEnabled() ? pl.getExperienceLevel() : 0);

        obj.setBayCount(roundToInt(std::sqrt(double(planet_defense))));
        if (base_fighters != 0)
            obj.addBays(5);

        obj.setTorpType(0);
        obj.setMass(100 + planet_defense + base_defense);
        obj.setShield(pl.getShield());
        if (planet_defense == 0 && base_defense == 0)
            obj.setShield(0);
    }
}

/** Unpack planet from VCR.
    \param obj      [in] Object
    \param pl       [in/out] Planet in simulator.
    \param orig_obj [in] Original object.
    \param opts     [in] Simulator options */
static void
unpackPlanet(const GVcrObject& obj, GSimPlanet& pl, const GVcrObject& orig_obj, const GSimOptions& opts)
{
    int fighters_lost = orig_obj.getFighterCount() - obj.getFighterCount();
    int integrity_remaining = 100 - obj.getDamage();
    if (integrity_remaining < 0)
        integrity_remaining = 0;

    if (opts.vcr_mode != opts.VcrHost && opts.vcr_mode != opts.VcrNuHost) {
        pl.setShield(obj.getShield());
        pl.setDamage(obj.getDamage());

        if (pl.getBaseBeamTech() > 0) {
            /* remove fighters */
            int new_base_fighters = pl.getBaseFighters() - fighters_lost;
            if (new_base_fighters < 0) {
                pl.setDefense(pl.getDefense() + new_base_fighters);
                pl.setBaseFighters(0);
            } else {
                pl.setBaseFighters(new_base_fighters);
            }

            /* remove torps */
            if (config.PlanetsHaveTubes() && config.UseBaseTorpsInCombat(pl.getOwner())) {
                int  torps_lost = orig_obj.getTorpCount() - obj.getTorpCount();
                long torp_cost = torps_lost * ::getTorp(obj.getTorpType()).getTorpCost();
                while (torp_cost > 0) {
                    bool did = false;
                    for (int i = 1; i <= NUM_TORPS; ++i) {
                        if (pl.getBaseTorps(i) > 0) {
                            pl.setBaseTorps(i, pl.getBaseTorps(i) - 1);
                            torp_cost -= ::getTorp(obj.getTorpType()).getTorpCost(); // FIXME? should be 'i'?
                            did = true;
                        }
                    }
                    if (!did)
                        break;
                }
            }
        }
    } else {
        pl.setShield(obj.getShield());

        /* reduce defense */
        pl.setDefense(rdivaddw(integrity_remaining * pl.getDefense(), 100, 0));

        if (pl.getBaseBeamTech() > 0) {
            /* remove fighters */
            int new_base_fighters = pl.getBaseFighters() - fighters_lost;
            if (new_base_fighters < 0) {
                pl.setDefense(pl.getDefense() + new_base_fighters);
                pl.setBaseFighters(0);
            } else {
                pl.setBaseFighters(new_base_fighters);
            }

            /* reduce equipment */
            pl.setBaseDefense(rdivaddw(integrity_remaining * pl.getBaseDefense(), 100, 0));
            pl.setBaseBeamTech(std::max(1, rdivaddw(integrity_remaining * pl.getBaseBeamTech(), 100, 0)));
            pl.setBaseTorpTech(std::max(1, rdivaddw(integrity_remaining * pl.getBaseTorpTech(), 100, 0)));

            /* add base damage */
            int new_damage = pl.getDamage() + obj.getDamage();
            if (new_damage > 100) {
                pl.setBaseBeamTech(0);
            } else {
                pl.setDamage(new_damage);
            }
        } else {
            /* remove fighters */
            pl.setDefense(pl.getDefense() - fighters_lost);
        }

        if (pl.getDefense() < 0)
            pl.setDefense(0);
    }
}

/** Handle a ship being killed.
    This implements the respawn logic for Squadrons.
    \return true iff ship respawns */
static bool
handleShipKilled(GSimShip& sh)
{
    if (sh.hasFunction(sf_Squadron) && sh.getBeamCount() > 1) {
        sh.setBeamCount(sh.getBeamCount() - 1);
        sh.setDamage(0);
        sh.setShield(100);
        return true;
    } else {
        sh.setOwner(0);
        return false;
    }
}

/** Make ship/ship VCR. This routine also does left/right randomisation.
    \param db    [in/out] Database
    \param left  [in/out] Left ship
    \param right [in/out] Right ship
    \param opts  [in] Simulation options
    \param type  [in] VCR Type
    \param result [in/out] Result meta-information (weight) */
static bool
makeShipShipVcr(GClassicVcrDatabase& db, GSimShip& left, GSimShip& right,
                const GSimOptions& opts,
                const GClassicVcrType type,
                const GPlayerArray<int>& level_base, GSimBattleResult& result)
{
    /* fight? */
    if (!isAttacking(left, right, opts) && !isAttacking(right, left, opts))
        return false;
    if (!(isArmed(left) || isArmed(right)))
        return false;

    /* swap them? */
    GSimShip *one, *two;
    bool swap_them = false;
    if (opts.random_left_right) {
        if (opts.seed_control) {
            swap_them = result.addSeries(2) == 0;
        } else {
            swap_them = (myRandom(2) == 0);
        }
    }
    if (swap_them)
        one = &right, two = &left;
    else
        one = &left, two = &right;

    /* set up fight */
    GClassicVcr vcr;
    vcr.setSeed(getSeed(opts, result));
    vcr.setSignature(0);
    packShip(vcr(0), *one, opts);
    packShip(vcr(1), *two, opts);
    applyShipModificators(vcr(0), false, *one, opts, level_base);
    applyShipModificators(vcr(1), false, *two, opts, level_base);
    applyOpponentModificators(vcr(0), *two, opts);
    applyOpponentModificators(vcr(1), *one, opts);

    /* left/right balance */
    if (opts.balancing_mode == opts.Balance360k) {
        if (vcr(1).getMass() > 140 && vcr(0).getBayCount() != 0) {
            if (opts.seed_control) {
                if (result.addSeries(2)) {
                    /* give bonus */
                    vcr(1).addMass(360);
                    result.this_battle_weight *= 59;
                    result.total_battle_weight *= 100;
                } else {
                    /* don't give bonus */
                    result.this_battle_weight *= 41;
                    result.total_battle_weight *= 100;
                }
            } else {
                if (myRandom(100) > 40)
                    vcr(1).addMass(360);
            }
        } else {
            /* The total_battle_weight for a series of simulations
               should depend only on the GSimOptions and the input
               situation. Due to random left/right, it can happen that
               for the same situation, we sometimes see the trigger
               for Balance360k, and sometimes we don't.

               Although GSimResultSummary can handle weights that
               change in the middle, it's better to use the clean way.
               A possible optimisation would apply this fix only if
               the reverse situation actually can trigger Balance360k. */
            if (opts.seed_control) {
                result.addSeries(2);
                result.this_battle_weight *= 50;
                result.total_battle_weight *= 100;
            }
        }
    } else if (opts.balancing_mode == opts.BalanceMasterAtArms) {
        /* Master-at-Arms bonus */
        applyMasterBonus(vcr, result, opts);
    }

    /* run it */
    uint16_t cap = type == vcr_PHost4 ? vcrcap_Beams | vcrcap_DeathRays | vcrcap_Experience : 0;
    Ptr<VcrPlayer> player = GClassicVcrEntry::createPlayerForType(type);
    player->setCapabilities(cap);
    player->checkVcr(vcr);

    GClassicVcrEntry& e = db.addEntry(vcr);
    e.setType(type, cap);

    player->setVisualizer(new VcrNullVisualizer);
    player->initVcr(vcr);
    player->setStat(0, one->getStat());
    player->setStat(1, two->getStat());
    player->playFastForward();
    while (player->playCycle())
        ;
    e.setResultFromPlayer(*player);

    /* copy back */
    unpackShip(e.getResult()(0), *one, opts);
    unpackShip(e.getResult()(1), *two, opts);

    bool again = false;
    switch (e.getStatus()) {
     case VcrPlayer::VCRS_LEFT_DESTROYED:
        again = handleShipKilled(*one);
        break;
     case VcrPlayer::VCRS_RIGHT_DESTROYED:
        again = handleShipKilled(*two);
        break;
     case VcrPlayer::VCRS_LEFT_CAPTURED:
        one->setOwner(two->getOwner());
        one->setCrew(10);
        one->setAggressiveness(GSimShip::agg_Passive);
        break;
     case VcrPlayer::VCRS_RIGHT_CAPTURED:
        two->setOwner(one->getOwner());
        two->setCrew(10);
        two->setAggressiveness(GSimShip::agg_Passive);
        break;
     case VcrPlayer::VCRS_TIMEOUT:
        break;
     default:
        /* results such as mutual capture or kill */
        /* FIXME: it seems HOST allows mutual capture to swap ships */
        one->setOwner(0);
        two->setOwner(0);
        break;
    }

    one->getStat() = player->getStat(0);
    two->getStat() = player->getStat(1);
    one->getStat().num_fights++;
    two->getStat().num_fights++;
    return again;
}

/** Make ship/planet VCR.
    \param db    [in/out] VCR database
    \param left  [in/out] Ship
    \param right [in/out] Planet
    \param opts  [in] Simulation options
    \param type  [in] VCR Type
    \param result [in/out] Result meta-information (weight) */
static bool
makeShipPlanetVcr(GClassicVcrDatabase& db, GSimShip& left, GSimPlanet& right,
                  const GSimOptions& opts,
                  const GClassicVcrType type,
                  const GPlayerArray<int>& level_base, GSimBattleResult& result)
{
    /* fight? */
    if (!isAttacking(left, right, opts))
        return false;

    /* set up fight */
    GClassicVcr vcr;
    vcr.setSeed(getSeed(opts, result));
    vcr.setSignature(0);
    // vcr._planet_temp_code = 0; /* host places temperature here */
    packShip(vcr(0), left, opts);
    applyShipModificators(vcr(0), true, left, opts, level_base);

    packPlanet(vcr(1), right, opts);
    if (opts.balancing_mode == opts.BalanceMasterAtArms)
        applyMasterBonus(vcr, result, opts);

    applyPlanetModificators(vcr(1), right, level_base);

    /* run it */
    uint16_t cap = type == vcr_PHost4 ? vcrcap_Beams | vcrcap_DeathRays | vcrcap_Experience : 0;
    Ptr<VcrPlayer> player = GClassicVcrEntry::createPlayerForType(type);
    player->checkVcr(vcr);
    player->setCapabilities(cap);

    GClassicVcrEntry& e = db.addEntry(vcr);
    e.setType(type, cap);

    player->setVisualizer(new VcrNullVisualizer);
    player->initVcr(vcr);
    player->setStat(0, left.getStat());
    player->setStat(1, right.getStat());
    player->playFastForward();
    while (player->playCycle())
        ;
    e.setResultFromPlayer(*player);

    /* copy back */
    unpackShip(e.getResult()(0), left, opts);
    unpackPlanet(e.getResult()(1), right, vcr(1), opts);

    bool again = false;
    switch (e.getStatus()) {
     case VcrPlayer::VCRS_LEFT_DESTROYED:
        again = handleShipKilled(left);
        break;
     case VcrPlayer::VCRS_RIGHT_DESTROYED:
        right.setOwner(0);
        break;
     case VcrPlayer::VCRS_LEFT_CAPTURED:
        left.setOwner(right.getOwner());
        left.setCrew(10);
        left.setAggressiveness(GSimShip::agg_Passive);
        break;
     case VcrPlayer::VCRS_RIGHT_CAPTURED:
        right.setOwner(left.getOwner());
        right.setBaseBeamTech(0);
        right.setFCode("???");
        break;
     case VcrPlayer::VCRS_TIMEOUT:
        break;
     default:
        /* results such as mutual capture or kill */
        left.setOwner(0);
        right.setOwner(0);
        break;
    }

    left.getStat() = player->getStat(0);
    right.getStat() = player->getStat(1);
    left.getStat().num_fights++;
    right.getStat().num_fights++;
    return again;
}

/** Compute maximum experience levels of all Commander ships. Finds the most experienced
    commander ships of all players, and propagates them via alliances.
    \param state [in] Simulation state
    \param opts  [in] Simulation options (incl. alliance settings)
    \param result [out] Experience level of most experienced Commander ship effective for each player */
static void
computeMaximumExperienceLevels(const GSimState& state, const GSimOptions& opts, GPlayerArray<int>& result)
{
    GPlayerArray<int> tmp;
    tmp.setAll(0);
    result.setAll(0);
    if (opts.isExperienceEnabled()) {
        // Find maximum experience levels for all players
        for (GSimState::ship_index_t i = 0; i < state.getNumShips(); ++i) {
            const GSimShip& sh = state.getShip(i);
            if (tmp.isValidIndex(sh.getOwner())
                && (sh.getFlags() & GSimShip::fl_Deactivated) == 0
                && sh.hasFunction(sf_Commander)
                && sh.getExperienceLevel() > tmp[sh.getOwner()])
            {
                tmp[sh.getOwner()] = sh.getExperienceLevel();
            }
        }
        // Propagate levels, honoring alliances
        for (int i = 1; i <= NUM_OWNERS; ++i) {
            int level = 0;
            for (int ally = 1; ally <= NUM_OWNERS; ++ally) {
                if (ally == i || (opts.honor_alliances && opts.alliance_settings.get(ally, i))) {
                    if (level < tmp[ally])
                        level = tmp[ally];
                }
            }
            result[i] = level;
        }
    }
}

/** Generate simulation according to HOST rules.
    \param state   [in/out] Simulation state
    \param opts    [in] Simulation options
    \param result  [in/out] Result meta-information
    \param monitor Progress monitor
    \param type    [in] VCR Type */
static void
simulateHost(GSimState& state, const GSimOptions& opts,
             GSimBattleResult& result, ProgressMonitor& monitor,
             const GClassicVcrType type)
{
    Ptr<GClassicVcrDatabase> db = new GClassicVcrDatabase();
    result.battles = db;

    /* compute Commander level limits */
    GPlayerArray<int> level_base;
    computeMaximumExperienceLevels(state, opts, level_base);

    /* compute battle order */
    std::vector<GSimShip*> battle_order;
    for (GSimState::ship_index_t i = 0; i < state.getNumShips(); ++i)
        if ((state.getShip(i).getFlags() & GSimObject::fl_Deactivated) == 0)
            battle_order.push_back(&state.getShip(i));

    /* simulate intercept-attack. */
    std::sort(battle_order.begin(), battle_order.end(), sortByIdBackwards);
    for (std::vector<GSimShip*>::size_type interceptor = 0; interceptor < battle_order.size(); ++interceptor) {
        /* FIXME: this loop is 98% identical to the one in
           simulatePHost. The main difference is that we iterate over
           a vector<GSimShip*> whereas PHost uses a
           vector<GSimObject*>. */
        if (GSimShip* iship = dynamic_cast<GSimShip*>(battle_order[interceptor])) {
            GSimShip* target;
            if (iship->getInterceptId() != 0 &&
                (target = const_cast<GSimShip*>(state.getShipById(iship->getInterceptId()))) != 0
                && target != iship)
            {
                bool loop = true;
                while (loop) {
                    monitor.tick();
                    loop = makeShipShipVcr(*db, *target, *iship, opts, type, level_base, result);
                    if (db->getNumBattles() != 0 && opts.only_one_simulation)
                        return;
                }
            }
        }
    }

    /* simulate. Outer loop selects right ship, inner loop selects left ship. */
    std::sort(battle_order.begin(), battle_order.end(), sortByBattleOrderTHost);
    for (std::vector<GSimShip*>::size_type right = 0; right < battle_order.size(); ++right) {
        for (std::vector<GSimShip*>::size_type left = 0; left < battle_order.size(); ++left) {
            if (left != right) {
                bool loop = true;
                while (loop) {
                    monitor.tick();
                    loop = makeShipShipVcr(*db, *battle_order[left], *battle_order[right], opts, type, level_base, result);
                    if (db->getNumBattles() != 0 && opts.only_one_simulation)
                        return;
                }
            }
        }
    }

    /* simulate fight vs planet */
    if (state.hasPlanet()) {
        for (std::vector<GSimShip*>::size_type left = 0; left < battle_order.size(); ++left) {
            bool loop = true;
            while (loop) {
                monitor.tick();
                loop = makeShipPlanetVcr(*db, *battle_order[left], state.getPlanet(), opts, type, level_base, result);
                if (db->getNumBattles() != 0 && opts.only_one_simulation)
                    return;
            }
        }
    }
}

/** Generate simulate according to PHost rules
    \param state   [in/out] Simulation state
    \param opts    [in] Simulation options
    \param result  [in/out] Result meta-information
    \param monitor Progress monitor
    \param type    [in] Exact type */
static void
simulatePHost(GSimState& state, const GSimOptions& opts, GSimBattleResult& result, ProgressMonitor& monitor, GClassicVcrType type)
{
    /* Quick&Dirty First Implementation

       This implementation is 99% identical to the one in ccbsim.
       Following features are missing:
       - PHost has *asymmetric* battle order rules
       - 'Commander' ships not implemented */
    Ptr<GClassicVcrDatabase> db = new GClassicVcrDatabase();
    result.battles = db;

    /* compute Commander level limits */
    GPlayerArray<int> level_base;
    computeMaximumExperienceLevels(state, opts, level_base);

    /* prepare planet */
    if (state.hasPlanet()) {
        GSimPlanet& p = state.getPlanet();
        p.setShield(100);
        p.setDamage(0);
    }

    /* compute battle order */
    std::vector<GSimObject*> battle_order;
    for (GSimState::ship_index_t i = 0; i < state.getNumShips(); ++i)
        if ((state.getShip(i).getFlags() & GSimObject::fl_Deactivated) == 0)
            battle_order.push_back(&state.getShip(i));
    if (state.hasPlanet())
        battle_order.push_back(&state.getPlanet());
    std::sort(battle_order.begin(), battle_order.end(), sortByBattleOrderPHost);

    /* simulate intercept-attack. */
    for (std::vector<GSimObject*>::size_type interceptor = 0; interceptor < battle_order.size(); ++interceptor) {
        if (GSimShip* iship = dynamic_cast<GSimShip*>(battle_order[interceptor])) {
            GSimShip* target;
            if (iship->getInterceptId() != 0 &&
                (target = const_cast<GSimShip*>(state.getShipById(iship->getInterceptId()))) != 0
                && target != iship)
            {
                bool loop = true;
                while (loop) {
                    monitor.tick();
                    loop = makeShipShipVcr(*db, *iship, *target, opts, type, level_base, result);
                    if (db->getNumBattles() != 0 && opts.only_one_simulation)
                        goto out;
                }
            }
        }
    }

    /* simulate. Outer loop picks aggressor, inner loop picks opponent */
    for (std::vector<GSimShip*>::size_type right = 0; right < battle_order.size(); ++right) {
        for (std::vector<GSimShip*>::size_type left = 0; left < battle_order.size(); ++left) {
            if (left != right) {
                bool loop = true;
                while (loop) {
                    loop = false;
                    monitor.tick();
                    if (GSimShip* lship = dynamic_cast<GSimShip*>(battle_order[left])) {
                        if (GSimShip* rship = dynamic_cast<GSimShip*>(battle_order[right])) {
                            loop = makeShipShipVcr(*db, *lship, *rship, opts, type, level_base, result);
                        } else if (GSimPlanet* rplan = dynamic_cast<GSimPlanet*>(battle_order[right])) {
                            loop = makeShipPlanetVcr(*db, *lship, *rplan, opts, type, level_base, result);
                        }
                    } else if (GSimPlanet* lplan = dynamic_cast<GSimPlanet*>(battle_order[left])) {
                        if (GSimShip* rship = dynamic_cast<GSimShip*>(battle_order[right])) {
                            loop = makeShipPlanetVcr(*db, *rship, *lplan, opts, type, level_base, result);
                        }
                    }
                    if (db->getNumBattles() != 0 && opts.only_one_simulation)
                        goto out;
                }
            }
        }
    }

 out:
    /* postprocess planet */
    if (state.hasPlanet()) {
        GSimPlanet& p = state.getPlanet();
        p.setDefense(p.getDefense() * (100L - p.getDamage()) / 100);
        if (p.getBaseBeamTech() > 0) {
            int base_damage = p.getBaseDamage() + p.getDamage();
            if (base_damage >= 100) {
                p.setBaseBeamTech(0); // delete base
            } else {
                p.setBaseDamage(base_damage);
                p.setBaseDefense(p.getBaseDefense() * (100L - p.getDamage()) / 100);
                p.setBaseBeamTech(getDamageTech(p.getBaseBeamTech(), p.getDamage()));
                p.setBaseTorpTech(getDamageTech(p.getBaseTorpTech(), p.getDamage()));
            }
        }
    }
}

#ifdef CONF_FLAK_SUPPORT
/********************************** FLAK *********************************/

// FIXME: ShipInfo and PermutedSorter as well as the following functions
// are almost copied verbatim from pdk-host.cc!

/** Get damage-restricted tech level for a base. */
static int
getBaseDamageTech(const GSimPlanet& pl, int have_tech)
{
    return std::min(int((100 - pl.getBaseDamage()) / 10),
                    int(have_tech));
}

/** Compute number of beams on a planet. */
static int
getPlanetBeamCount(const GSimPlanet& pl)
{
    int defense = pl.getDefense();
    if (pl.hasBase())
        defense += pl.getBaseDefense();
    defense = roundToInt(std::sqrt(defense / 3.0));
    return std::min(defense, int(config.AllowAlternativeCombat() ? FLAK_MAX_BEAMS : 10));
}

/** Compute beam type on a planet. */
static int
getPlanetBeamTech(const GSimPlanet& pl)
{
    int tech = roundToInt(std::sqrt(pl.getDefense() / 2.0));
    if (tech > 10)
        return 10;
    else if (pl.hasBase() && getBaseDamageTech(pl, pl.getBaseBeamTech()) > tech)
        return getBaseDamageTech(pl, pl.getBaseBeamTech());
    else
        return tech;
}

/** Compute number of planetary tubes. */
static int
getPlanetTubeCount(const GSimPlanet& pl)
{
    if (!config.PlanetsHaveTubes())
        return 0;
    int defense = pl.getDefense();
    if (pl.hasBase())
        defense += pl.getBaseDefense();
    defense = roundToInt(std::sqrt(defense / 4.0));
    return std::min(defense, int(FLAK_MAX_TORPS));
}

/** Compute torpedo type of a planet. */
static int
getPlanetTorpType(const GSimPlanet& pl)
{
    int tech = roundToInt(std::sqrt(pl.getDefense() / 2.0));
    if (tech > NUM_TORPS)
        return NUM_TORPS;
    else if (pl.hasBase() && getBaseDamageTech(pl, pl.getBaseTorpTech()) > tech)
        return getBaseDamageTech(pl, pl.getBaseTorpTech());
    else
        return tech;
}

/** Compute number of torpedoes on a planet. */
static int
getPlanetTorpCount(const GSimPlanet& pl)
{
    int torps = getPlanetTubeCount(pl) * config.PlanetaryTorpsPerTube(pl.getOwner());
    if (pl.hasBase())
        torps += pl.getBaseTorpsAsType(getPlanetTorpType(pl));
    return torps;
}

/** Compute number of fighter bays on a planet. */
static int
getPlanetBayCount(const GSimPlanet& pl)
{
    int bays = roundToInt(std::sqrt((double) pl.getDefense()));
    if (pl.hasBase())
        bays += 5;
    return bays;
}

/** Compute number of fighters on a planet. */
static int
getPlanetFighterCount(const GSimPlanet& pl)
{
    int fighters = roundToInt(std::sqrt((double) pl.getDefense()));
    if (pl.hasBase())
        fighters += pl.getBaseFighters();
    return fighters;
}

/** Compute combat mass of a planet. */
static int
getPlanetCombatMass(const GSimPlanet& pl)
{
    int mass = 100 + pl.getDefense();
    if (pl.hasBase())
        mass += pl.getBaseDefense();
    return mass;
}


namespace {
    struct ShipInfo {
        /** FCBO value, plus 100. */
        int fcbo_plus_100;
        /** Ship data. */
        TFlakShip data;
        /** Link to original unit. */
        GSimObject* orig;

        ShipInfo& initFromShip(GSimShip& sh);
        ShipInfo& initFromPlanet(GSimPlanet& pl);

        /** True iff this is a planet. */
        bool  isPlanet() const
            { return data.flags & flak_IsPlanet; }
    };
}

/** Initialize data for a ship. */
ShipInfo&
ShipInfo::initFromShip(GSimShip& sh)
{
    orig = &sh;

    const int level = sh.getExperienceLevel();

    storeBasicString(data.name, convertUtf8ToGame(sh.getName()));
    data.damage_init   = sh.getDamage();
    data.crew_init     = sh.getCrew();
    data.id            = sh.getId();
    data.player        = sh.getOwner();
    data.hull          = sh.getHull();
    data.level         = level;
    data.beam_count    = sh.getBeamCount();
    data.beam_type     = sh.getBeamType();
    data.torp_lcount   = sh.getTorpLauncherCount();
    data.torp_count    = sh.getTorpLauncherCount() ? sh.getAmmo() : 0;
    data.torp_type     = sh.getTorpType();
    data.bay_count     = sh.getBayCount();
    data.fighter_count = sh.getTorpLauncherCount() ? 0 : sh.getAmmo();
    data.mass          = sh.getMass();
    data.shield_init   = sh.getShield();

    /* NTP */
    if (sh.getFCode() == "NTP")
        data.fighter_count = data.torp_count = 0;

    /* ESB */
    int esb = 0;
    if (config.AllowEngineShieldBonus())
        esb += config.EngineShieldBonusRate(data.player);
    if (level)
        esb += config.EModEngineShieldBonusRate(level);
    if (esb)
        data.mass += getEngine(sh.getEngineType()).getCost().get(el_Money) * esb / 100;

    /* Fed crew bonus */
    if (config.AllowFedCombatBonus() && raceId(data.player) == 1)
        data.mass += 50;

    /* extra bays */
    if (data.bay_count) {
        data.bay_count += config.ExtraFighterBays(data.player);
        data.bay_count += getExperienceBonus(config.EModExtraFighterBays, level);
        if (data.bay_count > FLAK_MAX_BAYS)
            data.bay_count = FLAK_MAX_BAYS;
    }

    data.flags = 0;  // not a planet
    fcbo_plus_100 = std::min(getFCodeValuePHost(sh) + 100, 1099);

    FlakBattle::initShip(data);

    return *this;
}

/** Initialize data from planet. */
ShipInfo&
ShipInfo::initFromPlanet(GSimPlanet& pl)
{
    orig = &pl;

    const int level = pl.getExperienceLevel();

    storeBasicString(data.name, convertUtf8ToGame(pl.getName()));
    data.damage_init   = 0; /* planet starts with 0 damage in every turn */
    data.crew_init     = 0;
    data.id            = pl.getId();
    data.player        = pl.getOwner();
    data.hull          = 0;
    data.level         = level;
    data.beam_count    = getPlanetBeamCount(pl);
    data.beam_type     = getPlanetBeamTech(pl);
    data.torp_lcount   = getPlanetTubeCount(pl);
    data.torp_count    = getPlanetTorpCount(pl);
    data.torp_type     = getPlanetTorpType(pl);
    data.bay_count     = getPlanetBayCount(pl);
    data.fighter_count = getPlanetFighterCount(pl);
    data.mass          = getPlanetCombatMass(pl);
    data.shield_init   = 100; /* planet starts with 100 shield in every turn */

    /* extra bays */
    if (data.bay_count) {
        if (level)
            data.bay_count += config.EModExtraFighterBays(level);
        if (data.bay_count > FLAK_MAX_BAYS)
            data.bay_count = FLAK_MAX_BAYS;
    }

    data.flags = flak_IsPlanet;
    fcbo_plus_100 = std::min(getFCodeValuePHost(pl) + 100, 1099);
    FlakBattle::initShip(data);

    return *this;
}

/** Check whether any ship from me can attack any ship from them. */
static bool
canAttackThisFleet(const FlakBattle& battle, const FlakFleet& me, const FlakFleet& them,
                   const std::vector<ShipInfo>& info,
                   const GSimOptions& opts)
{
    // shortcut
    if (&me == &them)
        return false;

    // check it
    for (int my_index = 0; my_index < me.data.num_ships; ++my_index) {
        for (int their_index = 0; their_index < them.data.num_ships; ++their_index) {
            const FlakShip& my_ship = battle.getShipByNumber(me.data.first_ship + my_index);
            const FlakShip& their_ship = battle.getShipByNumber(them.data.first_ship + their_index);
            if (isAttackingAny(*info[me.data.first_ship + my_index].orig, *info[them.data.first_ship + their_index].orig, opts))
                if (my_ship.isArmed() || their_ship.isArmed())
                    return true;
        }
    }
    return false;
}

/** Compute attack list for one fleet. */
static void
computeAttackList(FlakBattle& battle, FlakFleet& fleet, const std::vector<ShipInfo>& info,
                  const GSimOptions& opts)
{
    int32 first_entry = battle.getAttackListNumber();

    /* only the FLAK_ALGORITHM >= 20060531 case */
    /* Fleet/fleet attack relations must be symmetrical. If any ship
       from a fleet can attack/be attacked by us, we must be allowed
       to attack all ships from that fleet. Otherwise, it could happen
       that the other fleet is still aggressive to us, but we do no
       longer have someone to attack. */
    /* FIXME: this code is really ugly. And it is almost a duplicate
       of the same thing in pdk-host.cc */
    for (int other_fleet_nr = 0; other_fleet_nr < battle.getNumFleets(); ++other_fleet_nr) {
        const FlakFleet& other_fleet = battle.getFleetByNumber(other_fleet_nr);
        if (canAttackThisFleet(battle, fleet, other_fleet, info, opts)) {
            for (int other_index = 0; other_index < other_fleet.data.num_ships; ++other_index) {
                const int other_member = other_index + other_fleet.data.first_ship;
                const FlakShip& them = battle.getShipByNumber(other_member);
                bool can_attack = false, match_pe = false, match_fc = false;
                // for each fleet member
                for (int fleet_index = 0; fleet_index < fleet.data.num_ships; ++fleet_index) {
                    const int fleet_member = fleet_index + fleet.data.first_ship;
                    const FlakShip& me = battle.getShipByNumber(fleet_member);
                    if (isAttackingAny(*info[fleet_member].orig, *info[other_member].orig, opts)) {
                        if (me.isArmed() || them.isArmed()) {
                            can_attack = true;
                            if (GSimShip* sh = dynamic_cast<GSimShip*>(info[fleet_member].orig))
                                if (sh->getAggressiveness() == them.data.player)
                                    match_pe = true;
                        }
                    } else {
                        match_fc = true;
                    }
                }
                if (can_attack) {
                    // we can attack it regularily
                    int bonus = myRandom(flak_config.RatingRandomBonus);
                    if (match_pe)
                        bonus += flak_config.RatingPEBonus;
                    if (!match_fc)
                        bonus += flak_config.RatingFullAttackBonus;
                    if (!bonus)
                        bonus = 1;
                    battle.addAttackListEntry(other_member, bonus);
                } else {
                    // we cannot attack it, so give it priority 0
                    battle.addAttackListEntry(other_member, 0);
                }
            }
        }
    }
    fleet.data.att_list_pointer = first_entry;
    fleet.data.att_list_size    = battle.getAttackListNumber() - first_entry;
}


namespace {
    /** Predicate for sort: sort a list of ShipInfo into their combat
        order. Groups players together, randomly permuted. Upon
        construction, this computes a permutation and uses it later
        on. */
    class PermutedSorter {
        int player_map[NUM_OWNERS];
        int frob(int pid)
            {
                if (pid > 0 && pid <= NUM_OWNERS)
                    return player_map[pid-1];
                return pid;
            }
     public:
        PermutedSorter();
        bool operator()(const ShipInfo& l, const ShipInfo& r)
            {
                int dif = frob(l.data.player) - frob(r.data.player);
                if (dif)
                    return dif < 0;
                dif = l.fcbo_plus_100 - r.fcbo_plus_100;
                if (dif)
                    return dif < 0;
                dif = l.data.id - r.data.id;
                if (dif)
                    return dif < 0;
                return (r.data.flags & flak_IsPlanet) == 0;
            }
    };
}

/** Constructor. Generates a permutation. */
PermutedSorter::PermutedSorter()
{
    for (int i = 1; i <= NUM_OWNERS; ++i)
        player_map[i-1] = i;
    for (int i = 1; i < NUM_OWNERS; ++i) {
        int j = myRandom(i+1);
        std::swap(player_map[i], player_map[j]);
    }
}

/** Update statistics counters from FLAK fight.
    \param fsh [in] Result of FLAK fight
    \param obj [in/out] Unit from simulation setup to update */
static void
unpackFlakStat(const FlakShip& fsh, GSimObject& obj)
{
    VcrStatItem& stat = obj.getStat();
    stat.torps_hit += fsh.status.torps_hit;
    if (stat.min_fighters_aboard > fsh.status.min_fighters_aboard)
        stat.min_fighters_aboard = fsh.status.min_fighters_aboard;
    stat.num_fights++;
}

/** Update simulation from FLAK ship.
    \param fsh [in] Result of FLAK fight for a ship
    \param ssh [in/out] Ship from simulation setup to update */
static void
unpackFlakShip(const FlakShip& fsh, GSimShip& ssh)
{
    ssh.setShield(roundToInt(fsh.status.shield));
    ssh.setDamage(roundToInt(fsh.status.damage));
    ssh.setCrew(roundToInt(fsh.status.crew));
    if (fsh.data.torp_lcount)
        ssh.setAmmo(fsh.status.torp_count);
    else
        ssh.setAmmo(fsh.status.fighter_count);
    if (fsh.data.ending_status < 0) {
        // died
        ssh.setAggressiveness(ssh.agg_Passive);
        ssh.setOwner(0);
    } else if (fsh.data.ending_status != 0 && fsh.data.ending_status != fsh.data.player) {
        // captured
        ssh.setOwner(fsh.data.ending_status);
        ssh.setCrew(10);
        ssh.setAggressiveness(0);
    }
    unpackFlakStat(fsh, ssh);
}

/** Update simulation from FLAK planet.
    \param fsh [in] Result of FLAK fight for a planet
    \param spl [in/out] Planet from simulation setup to update */
static void
unpackFlakPlanet(const FlakShip& fsh, GSimPlanet& spl)
{
    spl.setDamage(roundToInt(fsh.status.damage));
    spl.setShield(roundToInt(fsh.status.shield));
    if (spl.hasBase() && spl.getDamage() >= 100)
        spl.setBaseBeamTech(0); /* remove the base */

    if (spl.hasBase()) {
        /* base fighters */
        int fighters_lost = fsh.data.fighter_count - fsh.status.fighter_count;
        int new_sbf = spl.getBaseFighters() - fighters_lost;
        if (new_sbf < 0)
            spl.setBaseFighters(0);
        else
            spl.setBaseFighters(new_sbf);

        /* reduce tech */
        int max_tech = (100 - spl.getDamage()) / 10;
        if (max_tech <= 0)
            max_tech = 1;
        if (spl.getBaseBeamTech() > max_tech)
            spl.setBaseBeamTech(max_tech);
        if (spl.getBaseTorpTech() > max_tech)
            spl.setBaseTorpTech(max_tech);
    }

    int torps_lost = fsh.data.torp_count - fsh.status.torp_count;
    // Inc(PlanetaryTorpsFired, lost);
    if (torps_lost > 0 && spl.hasBase()) {
        long total_cost = torps_lost * getTorp(fsh.data.torp_type).getTorpCost();
        while (total_cost > 0) {
            bool did = 0;
            for (int i = 1; i <= NUM_TORPS; ++i) {
                if (spl.getBaseTorps(i) > 0 && getTorp(i).getTorpCost() <= total_cost) {
                    spl.setBaseTorps(i, spl.getBaseTorps(i) - 1);
                    total_cost -= getTorp(i).getTorpCost();
                    did = true;
                }
            }
            if (!did)
                total_cost = 0;
        }
    }

    if (fsh.data.ending_status != 0 && fsh.data.ending_status != fsh.data.player) {
        if (fsh.data.ending_status < 0)
            spl.setOwner(0);
        else
            spl.setOwner(fsh.data.ending_status);
        // spl.setAggressiveness(spl.agg_Passive);
        spl.setDefense(0);
        spl.setBaseBeamTech(0);
        spl.setFCode("???");
        spl.setShield(0);
    }

    unpackFlakStat(fsh, spl);
}

static void
simulateFLAK(GSimState& state, const GSimOptions& opts, GSimBattleResult& result, ProgressMonitor& monitor)
{
    Ptr<GFlakVcrDatabase> db = new GFlakVcrDatabase();
    result.battles = db;

    /* compute list of ships */
    std::vector<ShipInfo> ships;
    for (GSimState::ship_index_t i = 0; i < state.getNumShips(); ++i)
        ships.push_back(ShipInfo().initFromShip(state.getShip(i)));
    if (state.hasPlanet())
        ships.push_back(ShipInfo().initFromPlanet(state.getPlanet()));

    if (ships.size() < 2)
        return;

    /* group by owner, using a random permutation of owners */
    std::sort(ships.begin(), ships.end(), PermutedSorter());

    /* count players */
    GPlayerSet players;
    int player_count = 0;
    for (size_t i = 0; i < ships.size(); ++i) {
        if (!players.contains(ships[i].data.player)) {
            players |= ships[i].data.player;
            ++player_count;
        }
    }

    if (player_count < 2)
        return;

    /* Now build fleets. */
    FlakBattle battle;
    int  cur_player = 0;
    int  cur_fcbo   = 0;
    bool cur_planet = false;
    int  cur_fleet_size = 0;
    for (FlakBattle::ship_t i = 0; i < ships.size(); ++i) {
        if (i == 0 || cur_player != ships[i].data.player
            || cur_fcbo != ships[i].fcbo_plus_100 / 100
            || cur_planet != ships[i].isPlanet()
            || cur_fleet_size >= flak_config.MaximumFleetSize)
        {
            battle.addFleet(ships[i].data.player);
            cur_fcbo       = ships[i].fcbo_plus_100 / 100;
            cur_planet     = ships[i].isPlanet();
            cur_player     = ships[i].data.player;
            cur_fleet_size = 0;
        }
        FlakBattle::ship_t sid = battle.addShip(ships[i].data);
        ASSERT(sid == i);
        ++cur_fleet_size;
    }

    /* Now we have all the fleets, compute attack lists */
    for (int i = 0; i < battle.getNumFleets(); ++i)
        computeAttackList(battle, battle.getFleetByNumber(i), ships, opts);

    /* Compute speeds, etc. */
    battle.initAfterSetup();

    /* Set random seed */
    battle.setInitialSeed(myRandom());

    if (battle.getNumFleets() == 0)
        return;

    /* run it... */
    monitor.tick();
    FlakNullVisualizer nv(battle);
    battle.init();
    while (!battle.playOneTick())
        ;

    /* evaluate */
    // ASSERT(ships.size() == battle.getNumShips()); no longer holds!!!
    for (FlakBattle::ship_t i = 0; i < state.getNumShips(); ++i) {
        FlakShip* pfsh = battle.getShipById(state.getShip(i).getId(), false);
        if (pfsh) {
            if (pfsh->isAlive()) {
                /* survived */
                pfsh->data.ending_status = 0;
            } else {
                /* captured or died */
                const FlakShip* captor = battle.findCaptor(*pfsh);
                int limit = (raceId(pfsh->data.player) == 2
                             && captor != 0
                             && raceId(captor->data.player) == 2) ? 150 : 99;
                if (captor && pfsh->status.crew < 0.5 && roundToInt(pfsh->status.damage) <= limit)
                    pfsh->data.ending_status = captor->data.player;
                else
                    pfsh->data.ending_status = -1;
            }
            unpackFlakShip(*pfsh, state.getShip(i));
        }
    }

    if (state.hasPlanet()) {
        FlakShip* pfsh = battle.getShipById(state.getPlanet().getId(), true);
        if (pfsh) {
            if (pfsh->isAlive()) {
                /* survived */
                pfsh->data.ending_status = 0;
            } else {
                /* captured or died */
                const FlakShip* captor = battle.findCaptor(*pfsh);
                if (captor)
                    pfsh->data.ending_status = captor->data.player;
                else
                    pfsh->data.ending_status = -1;
            }
            unpackFlakPlanet(*pfsh, state.getPlanet());
        }
    }

    /* add battle to VCR db */
    db->addBattle(battle);
}
#endif

/******************************* Main Entry ******************************/

/** Simulate once. Takes this GSimState as input and updates it
    accordingly. Caller can then evaluate the GSimState and the GSimBattleResult.
    \param state     [in/out] Ships to simulate
    \param opts      [in] Simulation options
    \param result    [in/out] Result
    \param monitor   Progress monitor */
void
runSimulation(GSimState& state, const GSimOptions& opts, GSimBattleResult& result, ProgressMonitor& monitor)
{
    if (opts.randomize_fcodes_on_every_fight)
        state.assignRandomFCodes();
    state.initializeStats();
    switch (opts.vcr_mode) {
     case GSimOptions::VcrHost:
        simulateHost(state, opts, result, monitor, vcr_Host);
        break;
     case GSimOptions::VcrNuHost:
        simulateHost(state, opts, result, monitor, vcr_NuHost);
        break;
     case GSimOptions::VcrPHost2:
        simulatePHost(state, opts, result, monitor, vcr_PHost2);
        break;
     case GSimOptions::VcrPHost3:
        simulatePHost(state, opts, result, monitor, vcr_PHost3);
        break;
     case GSimOptions::VcrPHost4:
        simulatePHost(state, opts, result, monitor, vcr_PHost4);
        break;
     case GSimOptions::VcrFLAK:
#ifdef CONF_FLAK_SUPPORT
        simulateFLAK(state, opts, result, monitor);
#endif
        break;
     default:
        ASSERT(0);
        break;
    }
}
