/**
  *  \file gfx/update.cc
  *  \brief Update Management
  *
  *  Implements a class GfxUpdateOptimizer, which can optimize
  *  GfxCanvas::update calls.
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "gfx/update.h"

/** Constructor.
    \param parent Canvas to actually draw on */
GfxUpdateOptimizer::GfxUpdateOptimizer(GfxCanvas& parent)
    : parent(parent), update_region()
{ }

/** Destructor. If there are outstanding updates, performs them. */
GfxUpdateOptimizer::~GfxUpdateOptimizer()
{
    doUpdates();
}

/** Perform outstanding updates. */
void
GfxUpdateOptimizer::doUpdates()
{
    parent.update(update_region);
    update_region = GfxRect();
}

/** Clear update region. The previous update region is returned in \c region,
    to give a strong hint to the caller not to discard it lighted-heartly. */
void
GfxUpdateOptimizer::clearUpdateRegion(GfxRect& region)
{
    region = update_region;
    update_region = GfxRect();
}

void
GfxUpdateOptimizer::drawHLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    parent.drawHLine(pt, npix, color, pat, alpha);
}

void
GfxUpdateOptimizer::drawVLine(const GfxPoint& pt, int npix, uint32 color, uint8 pat, alpha_type alpha)
{
    parent.drawVLine(pt, npix, color, pat, alpha);
}

void
GfxUpdateOptimizer::drawPixel(const GfxPoint& pt, uint32 color, alpha_type alpha)
{
    parent.drawPixel(pt, color, alpha);
}

void
GfxUpdateOptimizer::drawBar(GfxRect rect, uint32 color, uint32 bg, const GfxFillPattern& pat, alpha_type alpha)
{
    parent.drawBar(rect, color, bg, pat, alpha);
}

void
GfxUpdateOptimizer::blitSurface(const GfxPoint& pt, GfxCanvas& src, GfxRect rect)
{
    parent.blitSurface(pt, src, rect);
}

void
GfxUpdateOptimizer::blitPattern(GfxRect rect, const GfxPoint& pt, int bytes_per_line, const uint8* data, uint32 color, uint32 bg, alpha_type alpha)
{
    parent.blitPattern(rect, pt, bytes_per_line, data, color, bg, alpha);
}

bool
GfxUpdateOptimizer::lock()
{
    return parent.lock();
}

void
GfxUpdateOptimizer::unlock()
{
    return parent.unlock();
}

void
GfxUpdateOptimizer::update(GfxRect r)
{
    update_region.include(r);
}

GfxRect
GfxUpdateOptimizer::computeClipRect(GfxRect r)
{
    return parent.computeClipRect(r);
}

bool
GfxUpdateOptimizer::computeOffset(GfxCanvas& other, GfxPoint& pt)
{
    return this == &other
        || parent.computeOffset(other, pt);
}

GfxPoint
GfxUpdateOptimizer::getSize()
{
    return parent.getSize();
}

int
GfxUpdateOptimizer::getBitsPerPixel()
{
    return parent.getBitsPerPixel();
}

uint32
GfxUpdateOptimizer::getPixel(const GfxPoint& pt)
{
    return parent.getPixel(pt);
}
