/**
  *  \file io/zipstream.h
  *  \brief Zip (deflate) stream
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_IO_ZIPSTREAM_H
#define PCC_V2_IO_ZIPSTREAM_H

#include <zlib.h>
#include "io/streams.h"
#include "cpluslib/types.h"

/** Zip stream. Reads and inflates a deflate stream. Does not support
    seeking nor writing. This supports raw deflate as well as skipping
    gzip headers.

    Note that this implements gzip header skipping directly instead of
    relying on zlib to do it, because I build this on systems where the
    system zlib doesn't have it. It's not that hard anyway.

    \todo Might need to skip zlib headers as well (for networking) */
class ZipStream : public Stream {
    /** Underlying worker stream. */
    Stream& worker;

    /** zlib stream. */
    z_stream zs;

    /** I/O buffers. */
    char bufIn[4096];
    char bufOut[4096];
    Bytef* outPtr;

    /** Position. */
    fpos_t bytesRead;

    /** State. zlib has been initialized if and only if the state is Decompressing. */
    enum {
        CheckHeader,
        CheckExtraDataLength,
        CheckExtraData,
        CheckFileName,
        CheckComment,
        CheckHeaderCRC,
        FinishedHeader,
        Decompressing
    } state;
    bool endReached;

    /** gzip header. */
    uint8_t gzHeaderFlags;
    uint16_t gzExtraLength;

    void zlibError(int state);
    void refillInput();
    void consumeInput(std::size_t n);
    bool consumeString();
    bool processHeader();

 public:
    ZipStream(Stream& worker, bool gzip);
    ~ZipStream();

    std::size_t read(char* buffer, std::size_t max);
    std::size_t write(const char* buffer, std::size_t max);
    int getCapabilities() const;
    void seek(fpos_t pos);
    fpos_t getPos();
    fpos_t getSize();
    string_t getName();
};

#endif
