/**
  *  \file reshack/grid.cc
  *  \brief Resource Editor: Grid Widgets
  *
  *  (c) 2001-2017 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "reshack/grid.h"
#include "ui/uiglobal.h"
#include "gfx/bmfont.h"

/** Constructor.
    \param font Font to display
    \param id Widget Id */
RHCharacterGrid::RHCharacterGrid(const GfxBitmapFont& font, int id)
    : UIIconGrid(GfxPoint(0, 0), 1, 1, id),
      font(font)
{
    init();
}

RHCharacterGrid::~RHCharacterGrid()
{ }

/** Get currently selected character number. */
uint32_t
RHCharacterGrid::getCurrentCharacter()
{
    return firsts[getCurrentY()] + getCurrentX();
}

/** Move cursor to a character. If that character is not visible
    because it has no glyph, goes to a character nearby. */
void
RHCharacterGrid::setCurrentCharacter(uint32_t n)
{
    if (firsts.size()) {
        int line = 0, col = 0;
        while (line < getNumLines()-1 && n >= firsts[line+1])
            ++line;
        if (n >= firsts[line] && n < firsts[line] + uint32_t(getNumColumns()))
            col = n - firsts[line];
        setCurrentPos(col, line);
    }
}

/** Initialize this widget. */
void
RHCharacterGrid::init()
{
    // Figure out cell size
    int minX = 10, minY = 10;
    for (uint32_t i = 0, end = font.getCurrentCharacterLimit(); i < end; ++i) {
        if (const GfxBitmapGlyph* g = font.getGlyph(i)) {
            if (g->getWidth() > minX)
                minX = g->getWidth();
            if (g->getHeight() > minY)
                minY = g->getHeight();
        }
    }
    minX += 2;
    minY += 2;

    // Figure out grid size
    int width = 8;
    while (width < 32 && minX * width < ui_screen->getSize().x / 3)
        width *= 2;

    // Populate "firsts" array
    bool prevLine = false;
    for (uint32_t i = 0, end = font.getCurrentCharacterLimit(); i < end; i += width) {
        // Anything on this line?
        bool thisLine = false;
        for (int j = 0; j < width; ++j)
            if (font.getGlyph(i + j))
                thisLine = true;

        // Show this line if anything is on it, or on the previous one (spacer)
        if (thisLine || prevLine)
            firsts.push_back(i);
        prevLine = thisLine;
    }
    if (firsts.empty()) {
        firsts.push_back(32);
    }

    // Configure base class
    setGeometry(GfxPoint(minX, minY), width, firsts.size());
}

void
RHCharacterGrid::drawIcon(GfxCanvas& can, const GfxRect& pos, int x, int y, FocusState state)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    if (state == NoFocus) {
        drawBackground(ctx, pos);
        ctx.setColor(UIColor::tc_Static);
    } else if (state == PrimaryFocus) {
        drawSolidBar(ctx, pos, UIColor::tc_Static);
        ctx.setColor(UIColor::tc_InvStatic);
    } else {
        drawBackground(ctx, pos);
        ctx.setColor(UIColor::tc_White);
    }
    if (const GfxBitmapGlyph* g = font.getGlyph(firsts[y] + x)) {
        g->draw(ctx, GfxPoint(pos.x + (pos.w - g->getWidth())/2, pos.y + 1));
    } else {
        ctx.setColor(UIColor::tc_Faded);
        drawLine(ctx, GfxPoint(pos.x + 1, pos.y + 1),         GfxPoint(pos.x + pos.w - 1, pos.y + pos.h - 2));
        drawLine(ctx, GfxPoint(pos.x + 1, pos.y + pos.h - 2), GfxPoint(pos.x + pos.w - 1, pos.y + 1));
    }
}
